import pytest
from amqp.exceptions import NotFound
from mock import Mock, patch
from six.moves import queue

from nameko.amqp import UndeliverableMessage
from nameko.events import event_handler
from nameko.standalone.events import event_dispatcher, get_event_exchange
from nameko.testing.services import entrypoint_waiter


handler_called = Mock()


class Service(object):
    name = 'destservice'

    @event_handler('srcservice', 'testevent')
    def handler(self, msg):
        handler_called(msg)


def test_dispatch(container_factory, rabbit_config):

    container = container_factory(Service)
    container.start()

    msg = "msg"

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handler', timeout=1):
        dispatch('srcservice', 'testevent', msg)
    handler_called.assert_called_once_with(msg)


class TestMandatoryDelivery(object):
    """ Test and demonstrate mandatory delivery.

    Dispatching an event should raise an exception when mandatory delivery
    is requested and there is no destination queue, as long as publish-confirms
    are enabled.
    """
    @pytest.fixture(autouse=True)
    def event_exchange(self, container_factory, rabbit_config):
        # use a service-based dispatcher to declare an event exchange
        container = container_factory(Service)
        container.start()

    def test_default(self, rabbit_config):
        # events are not mandatory by default;
        # no error when routing to a non-existent handler
        dispatch = event_dispatcher()
        dispatch("srcservice", "bogus", "payload")

    def test_mandatory_delivery(self, rabbit_config):
        # requesting mandatory delivery will result in an exception
        # if there is no bound queue to receive the message
        dispatch = event_dispatcher(mandatory=True)
        with pytest.raises(UndeliverableMessage):
            dispatch("srcservice", "bogus", "payload")

    def test_mandatory_delivery_no_exchange(self, rabbit_config):
        # requesting mandatory delivery will result in an exception
        # if the exchange does not exist
        dispatch = event_dispatcher(mandatory=True)
        with pytest.raises(NotFound):
            dispatch("bogus", "bogus", "payload")

    @patch('nameko.amqp.publish.warnings')
    def test_confirms_disabled(self, warnings, rabbit_config):
        # no exception will be raised if confirms are disabled,
        # even when mandatory delivery is requested,
        # but there will be a warning raised
        dispatch = event_dispatcher(mandatory=True, use_confirms=False)
        dispatch("srcservice", "bogus", "payload")
        assert warnings.warn.called


class TestConfigurability(object):
    """
    Test and demonstrate configuration options for the standalone dispatcher
    """

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.parametrize("parameter", [
        # delivery options
        'delivery_mode', 'mandatory', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy',
        # other arbitrary publish kwargs
        'correlation_id', 'user_id', 'bogus_param'
    ])
    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_regular_parameters(
        self, parameter, mock_container, producer
    ):
        """ Verify that most parameters can be specified at instantiation time.
        """
        value = Mock()

        dispatch = event_dispatcher(**{parameter: value})

        dispatch("service-name", "event-type", "event-data")
        assert producer.publish.call_args[1][parameter] == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_restricted_parameters(
        self, mock_container, producer
    ):
        """ Verify that providing routing parameters at instantiation
        time has no effect.
        """
        exchange = Mock()
        routing_key = Mock()

        dispatch = event_dispatcher(
            exchange=exchange, routing_key=routing_key)

        service_name = "service-name"
        event_exchange = get_event_exchange(service_name)
        event_type = "event-type"

        dispatch(service_name, event_type, "event-data")

        assert producer.publish.call_args[1]['exchange'] == event_exchange
        assert producer.publish.call_args[1]['routing_key'] == event_type


class TestSSL(object):

    @pytest.mark.usefixtures("rabbit_config")
    def test_event_dispatcher_over_ssl(
        self, container_factory, rabbit_ssl_uri, rabbit_ssl_options
    ):
        class Service(object):
            name = "service"

            @event_handler("service", "event")
            def echo(self, event_data):
                return event_data

        container = container_factory(Service)
        container.start()

        dispatch = event_dispatcher(uri=rabbit_ssl_uri, ssl=rabbit_ssl_options)

        with entrypoint_waiter(container, 'echo') as result:
            dispatch("service", "event", "payload")
        assert result.get() == "payload"



import os
from distutils import spawn

import pytest


def toxiproxy_is_installed():
    return spawn.find_executable('toxiproxy-server') is not None


def on_travis():
    return os.environ.get('TRAVIS', False)


def skip_toxiproxy():
    if not on_travis():
        return not toxiproxy_is_installed()


skip_if_no_toxiproxy = pytest.mark.skipif(
    skip_toxiproxy(), reason="toxiproxy not installed"
)



import uuid
from contextlib import contextmanager

import eventlet
import pytest
from eventlet.event import Event
from eventlet.semaphore import Semaphore
from greenlet import GreenletExit  # pylint: disable=E0611
from kombu import Exchange
from kombu.connection import Connection
from kombu.exceptions import OperationalError
from mock import Mock, call, patch
from six.moves import queue

from nameko import config
from nameko.amqp.consume import Consumer
from nameko.constants import (
    AMQP_URI_CONFIG_KEY, HEARTBEAT_CONFIG_KEY, MAX_WORKERS_CONFIG_KEY
)
from nameko.containers import WorkerContext
from nameko.events import event_handler
from nameko.exceptions import (
    ContainerBeingKilled, IncorrectSignature, MalformedRequest, MethodNotFound,
    RemoteError, ReplyQueueExpiredWithPendingReplies, UnknownService
)
from nameko.extensions import DependencyProvider
from nameko.rpc import (
    Client, ClusterRpc, ReplyListener, Responder, Rpc, RpcConsumer, ServiceRpc,
    rpc
)
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import dummy, entrypoint_hook, entrypoint_waiter
from nameko.testing.utils import get_extension
from nameko.testing.waiting import wait_for_call
from nameko.utils.retry import retry

from test import skip_if_no_toxiproxy


class ExampleError(Exception):
    pass


hello = object()
translations = {
    'en': {hello: 'hello'},
    'fr': {hello: 'bonjour'},
}


class Translator(DependencyProvider):

    def get_dependency(self, worker_ctx):
        def translate(value):
            lang = worker_ctx.data['language']
            return translations[lang][value]
        return translate


class LanguageContext(DependencyProvider):

    def get_dependency(self, worker_ctx):
        @contextmanager
        def set_language(language):
            original_language = worker_ctx.data.get('language')
            worker_ctx.data['language'] = language
            yield
            worker_ctx.data['language'] = original_language
        return set_language


class ExampleService(object):
    name = 'exampleservice'

    translate = Translator()
    language_context = LanguageContext()
    example_rpc = ServiceRpc('exampleservice')
    unknown_rpc = ServiceRpc('unknown_service')

    @rpc
    def task_a(self, *args, **kwargs):
        return "result_a"

    @rpc
    def task_b(self, *args, **kwargs):
        return "result_b"

    @rpc
    def call_async(self):
        res1 = self.example_rpc.task_a.call_async()
        res2 = self.example_rpc.task_b.call_async()
        res3 = self.example_rpc.echo.call_async()
        return [res2.result(), res1.result(), res3.result()]

    @rpc
    def call_unknown(self):
        return self.unknown_rpc.any_method()

    @rpc
    def echo(self, *args, **kwargs):
        return args, kwargs

    @rpc
    def say_hello(self):
        return self.translate(hello)

    @rpc
    def greet_france(self):
        with self.language_context('fr'):
            return self.example_rpc.say_hello()

    @event_handler('srcservice', 'eventtype')
    def async_task(self):
        pass  # pragma: no cover

    @rpc
    def raises(self):
        raise ExampleError("error")


@pytest.yield_fixture
def get_rpc_exchange():
    with patch('nameko.rpc.get_rpc_exchange', autospec=True) as patched:
        yield patched


@pytest.mark.usefixtures("memory_rabbit_config")
def test_rpc_consumer(get_rpc_exchange, mock_container):

    container = mock_container
    container.shared_extensions = {}
    container.service_name = "exampleservice"
    container.service_cls = Mock(rpcmethod=lambda: None)

    exchange = Exchange("some_exchange")
    get_rpc_exchange.return_value = exchange

    consumer = RpcConsumer().bind(container)

    entrypoint = Rpc().bind(container, "rpcmethod")
    entrypoint.rpc_consumer = consumer

    entrypoint.setup()
    consumer.setup()

    queue = consumer.queue
    assert queue.name == "rpc-exampleservice"
    assert queue.routing_key == "exampleservice.*"
    assert queue.exchange == exchange
    assert queue.durable

    consumer.register_provider(entrypoint)
    assert consumer._providers == set([entrypoint])

    routing_key = "exampleservice.rpcmethod"
    assert consumer.get_provider_for_method(routing_key) == entrypoint

    routing_key = "exampleservice.invalidmethod"
    with pytest.raises(MethodNotFound):
        consumer.get_provider_for_method(routing_key)

    consumer.unregister_provider(entrypoint)
    assert consumer._providers == set()


@pytest.mark.usefixtures("memory_rabbit_config")
def test_reply_listener(get_rpc_exchange, mock_container):

    container = mock_container
    container.shared_extensions = {}
    container.service_name = "exampleservice"

    exchange = Exchange("some_exchange")
    get_rpc_exchange.return_value = exchange

    reply_listener = ReplyListener().bind(container)

    forced_uuid = uuid.uuid4().hex

    with patch('nameko.rpc.uuid', autospec=True) as patched_uuid:
        patched_uuid.uuid4.return_value = forced_uuid

        reply_listener.setup()

        queue = reply_listener.queue
        assert queue.name == "rpc.reply-exampleservice-{}".format(forced_uuid)
        assert queue.exchange == exchange
        assert queue.routing_key == forced_uuid

    correlation_id = 1
    get_response = reply_listener.register_for_reply(correlation_id)
    assert correlation_id in reply_listener.pending

    reply_event = reply_listener.pending[correlation_id]

    payload = {"result": "msg", "error": None}
    message = Mock()
    message.properties.get.return_value = correlation_id
    reply_listener.handle_message(payload, message)

    assert reply_listener.pending == {}

    assert reply_event.ready()
    assert reply_event.wait() == payload
    assert get_response() == payload

    with patch('nameko.rpc._log', autospec=True) as log:
        reply_listener.handle_message("msg", message)
        assert log.debug.call_args == call(
            'Unknown correlation id: %s', correlation_id)


class TestClient(object):

    def test_getattr(self):
        publish, register_for_reply, context_data = Mock(), Mock(), Mock()

        client = Client(publish, register_for_reply, context_data)
        service_client = client.service
        assert client != service_client
        assert client.publish == service_client.publish
        assert client.register_for_reply == service_client.register_for_reply
        assert client.service_name is None
        assert client.method_name is None
        assert service_client.service_name == "service"
        assert service_client.method_name is None

        method1_client = service_client.method1
        assert service_client != method1_client
        assert service_client.publish == method1_client.publish
        assert service_client.register_for_reply == (
            method1_client.register_for_reply
        )
        assert service_client.service_name == "service"
        assert service_client.method_name is None
        assert method1_client.service_name == "service"
        assert method1_client.method_name == "method1"

        method2_client = service_client.method2

        method2_client = service_client.method2
        assert service_client != method2_client
        assert service_client.publish == method2_client.publish
        assert service_client.register_for_reply == (
            method2_client.register_for_reply
        )
        assert service_client.service_name == "service"
        assert service_client.method_name is None
        assert method2_client.service_name == "service"
        assert method2_client.method_name == "method2"

        assert method1_client != method2_client

        with pytest.raises(AttributeError):
            method1_client.attr
        with pytest.raises(AttributeError):
            method2_client.attr

    def test_identifier(self):
        publish, register_for_reply, context_data = Mock(), Mock(), Mock()

        client = Client(publish, register_for_reply, context_data)
        service_client = client.service
        method1_client = service_client.method1
        method2_client = service_client.method2

        assert client.identifier == "*.*"
        assert service_client.identifier == "service.*"
        assert method1_client.identifier == "service.method1"
        assert method2_client.identifier == "service.method2"

    def test_dict_access(self):
        publish, register_for_reply, context_data = Mock(), Mock(), Mock()

        client = Client(publish, register_for_reply, context_data)

        assert client['service'].identifier == "service.*"
        assert client['service']['method'].identifier == "service.method"

    def test_cannot_invoke_unspecified_client(self):
        publish, register_for_reply, context_data = Mock(), Mock(), Mock()

        client = Client(publish, register_for_reply, context_data)

        with pytest.raises(ValueError):
            client()

        with pytest.raises(ValueError):
            client.service()


# =============================================================================
# INTEGRATION TESTS
# =============================================================================


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_consumer_connections(
    container_factory, get_vhost, rabbit_manager
):

    class ExampleService(object):
        name = 'exampleservice'

        other_rpc = ServiceRpc('other')

        @rpc
        def echo(self, arg):
            return arg  # pragma: no cover

    container = container_factory(ExampleService)
    container.start()

    # we should have 2 queues:
    #   * RPC requests
    #   * RPC replies
    vhost = get_vhost(config['AMQP_URI'])
    queues = rabbit_manager.get_queues(vhost)
    assert len(queues) == 2

    # each one should have one consumer
    rpc_queue = rabbit_manager.get_queue(vhost, "rpc-exampleservice")
    assert len(rpc_queue['consumer_details']) == 1

    queue_names = [queue['name'] for queue in queues]
    reply_queue_names = [name for name in queue_names if 'rpc.reply' in name]
    assert len(reply_queue_names) == 1
    reply_queue_name = reply_queue_names[0]
    reply_queue = rabbit_manager.get_queue(vhost, reply_queue_name)
    assert len(reply_queue['consumer_details']) == 1

    # with one connection each
    # (connection sharing not supported until we migrate away from kombu)
    consumer_connection_names = set(
        queue['consumer_details'][0]['channel_details']['connection_name']
        for queue in [rpc_queue, reply_queue]
    )
    assert len(consumer_connection_names) == 2


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_args_kwargs(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with entrypoint_hook(container, 'echo') as echo:
        assert echo() == ((), {})
        assert echo("a", "b") == (("a", "b"), {})
        assert echo(foo="bar") == ((), {'foo': 'bar'})
        assert echo("arg", kwarg="kwarg") == (("arg",), {'kwarg': 'kwarg'})


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_context_data(container_factory):

    container = container_factory(ExampleService)
    container.start()

    context_data = {
        'language': 'en',
        'auth_token': '123456789'
    }

    with entrypoint_hook(container, 'say_hello', context_data) as say_hello:
        assert say_hello() == "hello"

    with entrypoint_hook(container, 'greet_france', context_data) as greet_france:
        assert greet_france() == "bonjour"

    context_data['language'] = 'fr'

    with entrypoint_hook(container, 'say_hello', context_data) as say_hello:
        assert say_hello() == "bonjour"


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("predictable_call_ids")
def test_rpc_headers(container_factory):

    container = container_factory(ExampleService)
    container.start()

    context_data = {
        'language': 'en',
        'otherheader': 'othervalue'
    }

    headers = {}

    rpc_consumer = get_extension(container, RpcConsumer)

    original_handle_result = rpc_consumer.handle_result

    def side_effect(msg, *args):
        original_handle_result(msg, *args)
        headers.update(msg.headers)

    with patch.object(rpc_consumer, 'handle_result') as patched:

        patched.side_effect = side_effect

        # use a standalone rpc client to call exampleservice.say_hello()
        with ServiceRpcClient("exampleservice", context_data) as client:
            assert client.say_hello() == "hello"

        # headers as per context data, plus call stack
        message = patched.call_args[0][0]
        assert message.headers == {
            'nameko.language': 'en',
            'nameko.otherheader': 'othervalue',
            'nameko.call_id_stack': ['standalone_rpc_client.0.0']
        }


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_existing_method(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with ServiceRpcClient("exampleservice") as client:
        assert client.task_a() == "result_a"
        assert client.task_b() == "result_b"


@pytest.mark.usefixtures("rabbit_config")
def test_async_rpc(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with entrypoint_hook(container, 'call_async') as call_async:
        assert call_async() == ["result_b", "result_a", [[], {}]]


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_incorrect_signature(container_factory):

    class Service(object):
        name = "service"

        @rpc
        def no_args(self):
            pass

        @rpc
        def args_only(self, a):
            pass

        @rpc
        def kwargs_only(self, a=None):
            pass

        @rpc
        def star_args(self, *args):
            pass

        @rpc
        def star_kwargs(self, **kwargs):
            pass

        @rpc
        def args_star_args(self, a, *args):
            pass

        @rpc
        def args_star_kwargs(self, a, **kwargs):
            pass

    container = container_factory(Service)
    container.start()

    method_calls = [
        (('no_args', (), {}), True),
        (('no_args', ('bad arg',), {}), False),
        (('args_only', ('arg',), {}), True),
        (('args_only', (), {'a': 'arg'}), True),
        (('args_only', (), {'arg': 'arg'}), False),
        (('kwargs_only', ('a',), {}), True),
        (('kwargs_only', (), {'a': 'arg'}), True),
        (('kwargs_only', (), {'arg': 'arg'}), False),
        (('star_args', ('a', 'b'), {}), True),
        (('star_args', (), {'c': 'c'}), False),
        (('star_kwargs', (), {'c': 'c'}), True),
        (('star_kwargs', ('a', 'b'), {}), False),
        (('args_star_args', ('a',), {}), True),
        (('args_star_args', ('a', 'b'), {}), True),
        (('args_star_args', (), {}), False),
        (('args_star_args', (), {'c': 'c'}), False),
        (('args_star_kwargs', ('a',), {}), True),
        (('args_star_kwargs', ('a', 'b'), {}), False),
        (('args_star_kwargs', ('a', 'b'), {'c': 'c'}), False),
        (('args_star_kwargs', (), {}), False),
    ]

    for signature, is_valid_call in method_calls:

        method_name, args, kwargs = signature

        with ServiceRpcClient("service") as client:
            method = getattr(client, method_name)

            if not is_valid_call:
                with pytest.raises(IncorrectSignature):
                    method(*args, **kwargs)
            else:
                method(*args, **kwargs)  # no raise


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_missing_method(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with ServiceRpcClient("exampleservice") as client:
        with pytest.raises(MethodNotFound) as exc_info:
            client.task_c()
    assert str(exc_info.value) == "task_c"


def test_rpc_invalid_message():
    entrypoint = Rpc()
    with pytest.raises(MalformedRequest) as exc:
        entrypoint.handle_message({'args': ()}, None)  # missing 'kwargs'
    assert 'Message missing `args` or `kwargs`' in str(exc)


@pytest.mark.usefixtures("rabbit_config")
def test_handle_message_raise_malformed_request(container_factory):
    container = container_factory(ExampleService)
    container.start()

    with pytest.raises(MalformedRequest):
        with patch('nameko.rpc.Rpc.handle_message') as handle_message:
            handle_message.side_effect = MalformedRequest('bad request')
            with ServiceRpcClient("exampleservice") as client:
                client.task_a()


@pytest.mark.usefixtures("rabbit_config")
def test_handle_message_raise_other_exception(container_factory):
    container = container_factory(ExampleService)
    container.start()

    with pytest.raises(RemoteError):
        with patch('nameko.rpc.Rpc.handle_message') as handle_message:
            handle_message.side_effect = Exception('broken')
            with ServiceRpcClient("exampleservice") as client:
                client.task_a()


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_method_that_raises(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with ServiceRpcClient("exampleservice") as client:
        with pytest.raises(RemoteError) as exc_info:
            client.raises()
    assert exc_info.value.exc_type == "ExampleError"


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_unknown_service(container_factory):
    container = container_factory(ExampleService)
    container.start()

    with ServiceRpcClient("exampleservice") as client:
        # success
        assert client.task_a()

        # failure
        with pytest.raises(RemoteError) as exc_info:
            client.call_unknown()

    assert exc_info.value.exc_type == "UnknownService"


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_unknown_service_standalone():

    with ServiceRpcClient("unknown_service") as client:
        with pytest.raises(UnknownService) as exc_info:
            client.anything()

    assert exc_info.value._service_name == 'unknown_service'


class TestContainerBeingKilled(object):

    @pytest.yield_fixture
    def service_rpc(self, rabbit_config):
        with ServiceRpcClient("service") as client:
            yield client

    @pytest.mark.usefixtures("rabbit_config")
    def test_container_killed(
        self, container_factory, service_rpc
    ):
        class Service(object):
            name = "service"

            @rpc
            def method(self):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        # check message is requeued if container throws ContainerBeingKilled
        with patch.object(container, 'spawn_worker') as spawn_worker:
            spawn_worker.side_effect = ContainerBeingKilled()

            with wait_for_call(Consumer, 'requeue_message'):
                service_rpc.method.call_async()


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_consumer_sharing(container_factory, rabbit_manager):
    """ Verify that the RpcConsumer stops when the first provider unregisters
    itself. Otherwise it would keep consuming messages for the unregistered
    provider, raising MethodNotFound.
    """

    container = container_factory(ExampleService)
    container.start()

    task_a = get_extension(container, Rpc, method_name="task_a")
    task_a_stop = task_a.stop

    task_b = get_extension(container, Rpc, method_name="task_b")
    task_b_stop = task_b.stop

    task_a_stopped = Event()

    def patched_task_a_stop():
        task_a_stop()  # stop immediately
        task_a_stopped.send(True)

    def patched_task_b_stop():
        eventlet.sleep(2)  # stop after 2 seconds
        task_b_stop()

    with patch.object(task_b, 'stop', patched_task_b_stop), \
            patch.object(task_a, 'stop', patched_task_a_stop):

        # stop the container and wait for task_a to stop
        # task_b will still be in the process of stopping
        eventlet.spawn(container.stop)
        task_a_stopped.wait()

        # try to call task_a.
        # should timeout, rather than raising MethodNotFound
        with ServiceRpcClient("exampleservice") as client:
            with pytest.raises(eventlet.Timeout):
                with eventlet.Timeout(1):
                    client.task_a()

    # kill the container so we don't have to wait for task_b to stop
    container.kill()


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_consumer_cannot_exit_with_providers(container_factory):
    container = container_factory(ExampleService)
    container.start()

    task_a = get_extension(container, Rpc, method_name="task_a")

    def never_stops():
        while True:
            eventlet.sleep()

    with patch.object(task_a, 'stop', never_stops):
        with pytest.raises(eventlet.Timeout):
            with eventlet.Timeout(1):
                container.stop()

    # kill off task_a's misbehaving rpc provider
    container.kill()


@pytest.mark.usefixtures("rabbit_config")
@patch('nameko.rpc.RPC_REPLY_QUEUE_TTL', new=100)
def test_reply_queue_removed_on_expiry(
    container_factory, get_vhost, rabbit_manager
):
    def list_queues():
        vhost = get_vhost(config['AMQP_URI'])
        return [
            queue['name']
            for queue in rabbit_manager.get_queues(vhost=vhost)
        ]

    class Service(object):
        name = "service"

        delegate_rpc = ServiceRpc('delegate')

        @dummy
        def method(self, arg):
            pass  # pragma: no cover

    container = container_factory(Service)
    container.start()

    reply_queue = [
        queue for queue in list_queues() if "rpc.reply" in queue
    ][0]

    container.stop()
    eventlet.sleep(0.15)  # sleep for >TTL
    assert reply_queue not in list_queues()


@pytest.mark.usefixtures("rabbit_config")
@patch('nameko.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_async_wait_longer_than_expiry(container_factory):
    """ ServiceRpc ReplyListener consumer runs in a thread and prevents the
    reply queue from expiring. Contrast to the single-threaded equivalent
    in nameko.standalone.rpc.
    """

    class Service(object):
        name = "service"

        delegate_rpc = ServiceRpc('delegate')

        @dummy
        def echo(self, arg):
            res = self.delegate_rpc.echo.call_async(arg)
            eventlet.sleep(0.4)  # sleep for 2x TTL
            return res.result()

    class DelegateService(object):
        name = "delegate"

        @rpc
        def echo(self, arg):
            return arg

    container = container_factory(Service)
    container.start()

    delegate_container = container_factory(DelegateService)
    delegate_container.start()

    with entrypoint_hook(container, 'echo') as echo:
        assert echo('foo') == 'foo'


@pytest.mark.usefixtures("rabbit_config")
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_request_longer_than_expiry(container_factory):

    class Service(object):
        name = "service"

        delegate_rpc = ServiceRpc('delegate')

        @dummy
        def echo(self, arg):
            return self.delegate_rpc.echo(arg)

    class DelegateService(object):
        name = "delegate"

        @rpc
        def echo(self, arg):
            eventlet.sleep(0.4)  # sleep for 2x TTL
            return arg

    container = container_factory(Service)
    container.start()

    delegate_container = container_factory(DelegateService)
    delegate_container.start()

    with entrypoint_hook(container, 'echo') as echo:
        assert echo('foo') == 'foo'


@pytest.mark.usefixtures("rabbit_config")
def test_cluster_rpc_dependency(container_factory):

    class Service(object):
        name = "service"

        cluster_rpc = ClusterRpc()

        @dummy
        def echo(self, arg):
            return self.cluster_rpc.delegate.echo(arg)

        @dummy
        def missing(self, arg):
            return self.cluster_rpc.missing.echo(arg)

    class DelegateService(object):
        name = "delegate"

        @rpc
        def echo(self, arg):
            return arg

    container = container_factory(Service)
    container.start()

    delegate_container = container_factory(DelegateService)
    delegate_container.start()

    with entrypoint_hook(container, 'echo') as echo:
        assert echo('foo') == 'foo'

    with entrypoint_hook(container, 'missing') as missing:
        with pytest.raises(UnknownService):
            missing("foo")


@pytest.mark.usefixtures('rabbit_config')
@skip_if_no_toxiproxy
class TestDisconnectedWhileWaitingForReply(object):  # pragma: no cover

    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            Consumer, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.yield_fixture
    def fast_expiry(self):
        with patch('nameko.rpc.RPC_REPLY_QUEUE_TTL', new=100):
            yield

    @pytest.yield_fixture
    def toxic_reply_listener(self, toxiproxy):
        with patch.object(ReplyListener, 'amqp_uri', new=toxiproxy.uri):
            yield

    @pytest.fixture(autouse=True)
    def container(
        self, container_factory, rabbit_manager, toxiproxy,
        fast_expiry, toxic_reply_listener
    ):

        def enable_after_queue_expires():
            eventlet.sleep(1)
            toxiproxy.enable()

        class Service(object):
            name = "service"

            delegate_rpc = ServiceRpc('delegate')

            @dummy
            def sleep(self):
                return self.delegate_rpc.sleep()

        class DelegateService(object):
            name = "delegate"

            @rpc
            def sleep(self):
                toxiproxy.disable()
                eventlet.spawn_n(enable_after_queue_expires)
                return "OK"

        # very fast heartbeat (2 seconds)
        with config.patch({HEARTBEAT_CONFIG_KEY: 2}):
            container = container_factory(Service)
            container.start()

            delegate_container = container_factory(DelegateService)
            delegate_container.start()

            yield container

    def test_reply_queue_removed_while_disconnected_with_pending_reply(
        self, container, toxiproxy
    ):
        with entrypoint_hook(container, 'sleep') as hook:
            with pytest.raises(ReplyQueueExpiredWithPendingReplies):
                hook()


@pytest.mark.usefixtures('rabbit_config')
@skip_if_no_toxiproxy
class TestReplyListenerDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.
    """
    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            Consumer, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.yield_fixture
    def toxic_reply_listener(self, toxiproxy):
        with patch.object(ReplyListener, 'amqp_uri', new=toxiproxy.uri):
            yield

    @pytest.fixture(autouse=True)
    def container(
        self, container_factory, toxiproxy, toxic_reply_listener
    ):
        class Service(object):
            name = "service"

            delegate_rpc = ServiceRpc('delegate')

            @dummy
            def echo(self, arg):
                return self.delegate_rpc.echo(arg)

        class DelegateService(object):
            name = "delegate"

            @rpc
            def echo(self, arg):
                return arg

        # very fast heartbeat (2 seconds)
        with config.patch({HEARTBEAT_CONFIG_KEY: 2}):
            container = container_factory(Service)
            container.start()

            delegate_container = container_factory(DelegateService)
            delegate_container.start()

            yield container

    def test_normal(self, container):
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg

    def test_down(self, container, toxiproxy):
        """ Verify we detect and recover from closed sockets.

        This failure mode closes the socket between the consumer and the
        rabbit broker.

        Attempting to read from the closed socket raises a socket.error
        and the connection is re-established.
        """
        reply_listener = get_extension(container, ReplyListener)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.enable()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.disable()

        # connection re-established
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg

    def test_upstream_timeout(self, container, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the consumer and the
        rabbit broker times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_upstream_blackhole` below.
        """
        reply_listener = get_extension(container, ReplyListener)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=100)

        # connection re-established
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg

    def test_upstream_blackhole(self, container, toxiproxy):
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the consumer to the
        rabbit broker is lost, but the socket remains open.

        Heartbeats sent from the consumer are not received by the broker. After
        two beats are missed the broker closes the connection, and subsequent
        reads from the socket raise a socket.error, so the connection is
        re-established.
        """
        reply_listener = get_extension(container, ReplyListener)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=0)

        # connection re-established
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg

    def test_downstream_timeout(self, container, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the rabbit broker and
        the consumer times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_downstream_blackhole` below, except that the consumer
        cancel will eventually (`timeout` milliseconds) raise a socket.error,
        which is ignored, allowing the teardown to continue.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        reply_listener = get_extension(container, ReplyListener)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=100)

        # connection re-established
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg

    def test_downstream_blackhole(
        self, container, toxiproxy
    ):  # pragma: no cover
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the rabbit broker to
        the consumer is lost, but the socket remains open.

        Heartbeat acknowledgements from the broker are not received by the
        consumer. After two beats are missed the consumer raises a "too many
        heartbeats missed" error.

        Cancelling the consumer requests an acknowledgement from the broker,
        which is swallowed by the socket. There is no timeout when reading
        the acknowledgement so this hangs forever.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        pytest.skip("skip until kombu supports recovery in this scenario")

        reply_listener = get_extension(container, ReplyListener)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=0)

        # connection re-established
        msg = "foo"
        with entrypoint_hook(container, 'echo') as echo:
            assert echo(msg) == msg


@pytest.mark.usefixtures('rabbit_config')
@skip_if_no_toxiproxy
class TestRpcConsumerDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.
    """
    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            Consumer, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.yield_fixture
    def toxic_rpc_consumer(self, toxiproxy):
        with patch.object(RpcConsumer, 'amqp_uri', new=toxiproxy.uri):
            yield

    @pytest.yield_fixture(autouse=True)
    def nontoxic_responder(self):
        def replacement_constructor(amqp_uri, *args, **kw):
            return Responder(config[AMQP_URI_CONFIG_KEY], *args, **kw)
        with patch('nameko.rpc.Responder', wraps=replacement_constructor):
            yield

    @pytest.yield_fixture
    def service_rpc(self):
        with ServiceRpcClient('service') as client:
            yield client

    @pytest.fixture
    def lock(self):
        return Semaphore()

    @pytest.fixture
    def tracker(self):
        return Mock()

    @pytest.yield_fixture(autouse=True)
    def container(
        self, container_factory, toxic_rpc_consumer, lock,
        tracker
    ):

        class Service(object):
            name = "service"

            @rpc
            def echo(self, arg):
                lock.acquire()
                lock.release()
                tracker(arg)
                return arg

        # very fast heartbeat (2 seconds)
        with config.patch({HEARTBEAT_CONFIG_KEY: 2}):
            container = container_factory(Service)
            container.start()

            # we have to let the container connect before disconnecting
            # otherwise we end up in retry_over_time trying to make the
            # initial connection; we get stuck there because it has a
            # function-local copy of "on_connection_error" that is never patched
            eventlet.sleep(.05)

            yield container

    def test_normal(self, container, service_rpc):
        assert service_rpc.echo("foo") == "foo"

    def test_down(self, container, service_rpc, toxiproxy):
        """ Verify we detect and recover from closed sockets.

        This failure mode closes the socket between the consumer and the
        rabbit broker.

        Attempting to read from the closed socket raises a socket.error
        and the connection is re-established.
        """
        rpc_consumer = get_extension(container, RpcConsumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.enable()
            return True

        with wait_for_call(
            rpc_consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.disable()

        # connection re-established
        assert service_rpc.echo("foo") == "foo"

    def test_upstream_timeout(self, container, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the consumer and the
        rabbit broker times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_upstream_blackhole` below.
        """
        rpc_consumer = get_extension(container, RpcConsumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            rpc_consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=100)

        # connection re-established
        assert service_rpc.echo("foo") == "foo"

    def test_upstream_blackhole(self, container, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the consumer to the
        rabbit broker is lost, but the socket remains open.

        Heartbeats sent from the consumer are not received by the broker. After
        two beats are missed the broker closes the connection, and subsequent
        reads from the socket raise a socket.error, so the connection is
        re-established.
        """
        rpc_consumer = get_extension(container, RpcConsumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            rpc_consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=0)

        # connection re-established
        assert service_rpc.echo("foo") == "foo"

    def test_downstream_timeout(self, container, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the rabbit broker and
        the consumer times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_downstream_blackhole` below, except that the consumer
        cancel will eventually (`timeout` milliseconds) raise a socket.error,
        which is ignored, allowing the teardown to continue.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        rpc_consumer = get_extension(container, RpcConsumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            rpc_consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=100)

        # connection re-established
        assert service_rpc.echo("foo") == "foo"

    def test_downstream_blackhole(
        self, container, service_rpc, toxiproxy
    ):  # pragma: no cover
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the rabbit broker to
        the consumer is lost, but the socket remains open.

        Heartbeat acknowledgements from the broker are not received by the
        consumer. After two beats are missed the consumer raises a "too many
        heartbeats missed" error.

        Cancelling the consumer requests an acknowledgement from the broker,
        which is swallowed by the socket. There is no timeout when reading
        the acknowledgement so this hangs forever.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        pytest.skip("skip until kombu supports recovery in this scenario")

        rpc_consumer = get_extension(container, RpcConsumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            rpc_consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=0)

        # connection re-established
        assert service_rpc.echo("foo") == "foo"

    def test_message_ack_regression(
        self, container, service_rpc, toxiproxy, lock, tracker
    ):
        """ Regression for https://github.com/nameko/nameko/issues/511
        """
        # prevent workers from completing
        lock.acquire()

        # fire entrypoint and block the worker;
        # break connection while the worker is active, then release worker
        with entrypoint_waiter(container, 'echo') as result:
            res = service_rpc.echo.call_async("msg1")
            while not lock._waiters:
                eventlet.sleep()  # pragma: no cover
            toxiproxy.disable()
            # allow connection to close before releasing worker
            eventlet.sleep(.1)
            lock.release()

        # entrypoint will return and reply will be received,
        # but the initiating message will not be ack'd
        assert result.get() == "msg1"
        assert res.result() == "msg1"

        # enabling connection will re-deliver the initiating message
        # and it will be processed again
        with entrypoint_waiter(container, 'echo') as result:
            toxiproxy.enable()
        assert result.get() == "msg1"

        # connection re-established, container should work again
        assert service_rpc.echo("msg2") == "msg2"


@pytest.mark.filterwarnings("ignore:Mandatory delivery:UserWarning")
@pytest.mark.usefixtures('rabbit_config')
@skip_if_no_toxiproxy
class TestClientDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.

    Publisher confirms must be enabled for some of these tests to pass. Without
    confirms, previously used but now dead connections will accept writes
    without raising. These tests are skipped in this scenario.

    Note that publisher confirms do not protect against sockets that remain
    open but do not deliver messages (i.e. `toxiproxy.set_timeout(0)`).
    This can only be mitigated with AMQP heartbeats (not yet supported)
    """

    @pytest.fixture(autouse=True)
    def server_container(self, container_factory):

        class Service(object):
            name = "server"

            @rpc
            def echo(self, arg):
                return arg

        container = container_factory(Service)
        container.start()

    @pytest.fixture(autouse=True)
    def client_container(self, container_factory, toxiproxy):

        class Service(object):
            name = "client"

            server_rpc = ServiceRpc("server", uri=toxiproxy.uri)

            @dummy
            def echo(self, arg):
                return self.server_rpc.echo(arg)

        container = container_factory(Service)
        container.start()
        return container

    @pytest.yield_fixture(autouse=True)
    def retry(self, request):
        retry = False
        if "publish_retry" in request.keywords:
            retry = True

        with patch.object(ServiceRpc.publisher_cls, 'retry', new=retry):
            yield

    @pytest.yield_fixture(params=[True, False])
    def use_confirms(self, request):
        with patch.object(
            ServiceRpc.publisher_cls, 'use_confirms', new=request.param
        ):
            yield request.param

    @pytest.mark.usefixtures('use_confirms')
    def test_normal(self, client_container):

        with entrypoint_hook(client_container, 'echo') as echo:
            assert echo(1) == 1
            assert echo(2) == 2

    @pytest.mark.usefixtures('use_confirms')
    def test_down(self, client_container, toxiproxy):
        with toxiproxy.disabled():

            with pytest.raises(OperationalError) as exc_info:
                with entrypoint_hook(client_container, 'echo') as echo:
                    echo(1)
            assert "ECONNREFUSED" in str(exc_info.value)

    @pytest.mark.usefixtures('use_confirms')
    def test_timeout(self, client_container, toxiproxy):
        with toxiproxy.timeout():

            with pytest.raises(OperationalError):
                with entrypoint_hook(client_container, 'echo') as echo:
                    echo(1)

    def test_reuse_when_down(self, client_container, toxiproxy):
        """ Verify we detect stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        with entrypoint_hook(client_container, 'echo') as echo:
            assert echo(1) == 1

        with toxiproxy.disabled():

            with pytest.raises(IOError):
                with entrypoint_hook(client_container, 'echo') as echo:
                    echo(2)

    def test_reuse_when_recovered(self, client_container, toxiproxy):
        """ Verify we detect and recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        with entrypoint_hook(client_container, 'echo') as echo:
            assert echo(1) == 1

        with toxiproxy.disabled():

            with pytest.raises(IOError):
                with entrypoint_hook(client_container, 'echo') as echo:
                    echo(2)

        with entrypoint_hook(client_container, 'echo') as echo:
            assert echo(3) == 3

    @pytest.mark.publish_retry
    def test_with_retry_policy(self, client_container, toxiproxy):
        """ Verify we automatically recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        with entrypoint_hook(client_container, 'echo') as echo:
            assert echo(1) == 1

        toxiproxy.disable()

        def enable_after_retry(args, kwargs, res, exc_info):
            toxiproxy.enable()
            return True

        # call 2 succeeds (after reconnecting via retry policy)
        with wait_for_call(
            Connection,
            "_establish_connection",
            callback=enable_after_retry
        ):
            with entrypoint_hook(client_container, 'echo') as echo:
                assert echo(2) == 2


@pytest.mark.filterwarnings("ignore:Mandatory delivery:UserWarning")
@pytest.mark.usefixtures('rabbit_config')
@skip_if_no_toxiproxy
class TestResponderDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.

    Publisher confirms must be enabled for some of these tests to pass. Without
    confirms, previously used but now dead connections will accept writes
    without raising. These tests are skipped in this scenario.

    Note that publisher confirms do not protect against sockets that remain
    open but do not deliver messages (i.e. `toxiproxy.set_timeout(0)`).
    This can only be mitigated with AMQP heartbeats (not yet supported)
    """

    @pytest.yield_fixture(autouse=True)
    def toxic_responder(self, toxiproxy):
        def replacement_constructor(amqp_uri, *args, **kwargs):
            return Responder(toxiproxy.uri, *args, **kwargs)
        with patch('nameko.rpc.Responder', wraps=replacement_constructor):
            yield

    @pytest.yield_fixture(params=[True, False])
    def use_confirms(self, request):
        value = request.param
        with patch.object(Responder.publisher_cls, 'use_confirms', new=value):
            yield value

    @pytest.yield_fixture(autouse=True)
    def retry(self, request):
        value = False
        if "publish_retry" in request.keywords:
            value = True

        with patch.object(Responder.publisher_cls, 'retry', new=value):
            yield

    @pytest.fixture
    def service_cls(self):
        class Service(object):
            name = "service"

            @rpc
            def echo(self, arg):
                return arg

        return Service

    @pytest.fixture(autouse=True)
    def container(self, container_factory, service_cls, rabbit_config):
        container = container_factory(service_cls)
        container.start()
        return container

    @pytest.yield_fixture
    def service_rpc(self):

        gts = []

        def kill_greenthreads():
            for gt in gts:
                try:
                    gt.kill()
                except GreenletExit:  # pragma: no cover
                    pass
            del gts[:]

        class ThreadSafeServiceRpc(object):
            """ The ServiceRpcClient is not thread-safe, so we can't create it
            in a fixture and use it directly from a new greenlet in a test
            (which we need to, because it hangs in many of these tests).

            Instead we fake the interface and make the underlying rpc call
            in a dedicated thread, ensuring it's never shared.
            """

            def __getattr__(self, name):
                def method(*args, **kwargs):
                    def call():
                        with ServiceRpcClient("service") as client:
                            return getattr(client, name)(*args, **kwargs)

                    gt = eventlet.spawn(call)
                    gts.append(gt)
                    return gt.wait()
                return method

            def abort(self):
                kill_greenthreads()

        client = ThreadSafeServiceRpc()
        yield client
        kill_greenthreads()

    @pytest.mark.usefixtures('use_confirms')
    def test_normal(self, service_rpc):
        assert service_rpc.echo(1) == 1
        assert service_rpc.echo(2) == 2

    @pytest.mark.usefixtures('use_confirms')
    def test_down(self, container, service_rpc, toxiproxy):
        with toxiproxy.disabled():

            eventlet.spawn_n(service_rpc.echo, 1)

            # the container will raise if the responder cannot reply
            with pytest.raises(OperationalError) as exc_info:
                container.wait()
            assert "ECONNREFUSED" in str(exc_info.value)

            service_rpc.abort()

    @pytest.mark.usefixtures('use_confirms')
    def test_timeout(self, container, service_rpc, toxiproxy):
        with toxiproxy.timeout():

            eventlet.spawn_n(service_rpc.echo, 1)

            # the container will raise if the responder cannot reply
            with pytest.raises(OperationalError):
                container.wait()

            service_rpc.abort()

    def test_reuse_when_down(self, container, service_rpc, toxiproxy):
        """ Verify we detect stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        with toxiproxy.disabled():

            eventlet.spawn_n(service_rpc.echo, 2)

            # the container will raise if the responder cannot reply
            with pytest.raises(IOError):
                container.wait()

            service_rpc.abort()

    def test_reuse_when_recovered(
        self, container, service_rpc, toxiproxy, container_factory,
        rabbit_config, service_cls
    ):
        """ Verify we detect and recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        with toxiproxy.disabled():

            eventlet.spawn_n(service_rpc.echo, 2)

            # the container will raise if the responder cannot reply
            with pytest.raises(IOError):
                container.wait()

        # create new container
        replacement_container = container_factory(service_cls)
        replacement_container.start()

        assert service_rpc.echo(3) == 3

    @pytest.mark.publish_retry
    def test_with_retry_policy(self, service_rpc, toxiproxy):
        """ Verify we automatically recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        toxiproxy.disable()

        def enable_after_retry(args, kwargs, res, exc_info):
            toxiproxy.enable()
            return True

        # call 2 succeeds (after reconnecting via retry policy)
        with wait_for_call(
            Connection,
            "_establish_connection",
            callback=enable_after_retry
        ):
            assert service_rpc.echo(2) == 2


class TestRpcConsumerConfigurability(object):

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_heartbeat(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = 999

        rpc_consumer = RpcConsumer(heartbeat=value).bind(mock_container)
        rpc_consumer.setup()

        assert rpc_consumer.consumer.connection.heartbeat == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_prefetch_count(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = 999

        rpc_consumer = RpcConsumer(prefetch_count=value).bind(mock_container)
        rpc_consumer.setup()

        assert rpc_consumer.consumer.prefetch_count == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_accept(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = ['yaml', 'json']

        rpc_consumer = RpcConsumer(accept=value).bind(mock_container)
        rpc_consumer.setup()

        assert rpc_consumer.consumer.accept == value


class TestReplyListenerConfigurability(object):

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_heartbeat(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = 999

        reply_listener = ReplyListener(heartbeat=value).bind(mock_container)
        reply_listener.setup()

        assert reply_listener.consumer.connection.heartbeat == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_prefetch_count(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = 999

        reply_listener = ReplyListener(
            prefetch_count=value
        ).bind(mock_container)
        reply_listener.setup()

        assert reply_listener.consumer.prefetch_count == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_accept(self, mock_container):
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        value = ['yaml', 'json']

        reply_listener = ReplyListener(accept=value).bind(mock_container)
        reply_listener.setup()

        assert reply_listener.consumer.accept == value


class TestConfigurability(object):
    """
    Test and demonstrate configuration options for the ServiceRpc
    """

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.parametrize("parameter", [
        # delivery options
        'delivery_mode', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy',
        # other arbitrary publish kwargs
        'user_id', 'bogus_param'
    ])
    def test_regular_parameters(
        self, parameter, mock_container, producer
    ):
        """ Verify that most parameters can be specified at ServiceRpc
        instantiation time.
        """
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.container = mock_container
        worker_ctx.context_data = {}

        value = Mock()

        service_rpc = ServiceRpc(
            "service-name", **{parameter: value}
        ).bind(mock_container, "service_rpc")

        service_rpc.reply_listener.setup()
        service_rpc.setup()

        service_rpc = service_rpc.get_dependency(worker_ctx)

        service_rpc.method.call_async()
        assert producer.publish.call_args[1][parameter] == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.usefixtures('predictable_call_ids')
    def test_headers(self, mock_container, producer):
        """ Headers can be provided at instantiation time, and are merged with
        Nameko headers.
        """
        mock_container.shared_extensions = {}
        mock_container.service_name = "service-name"

        # use a real worker context so nameko headers are generated
        service = Mock()
        entrypoint = Mock(method_name="method")
        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data={'context': 'data'}
        )

        nameko_headers = {
            'nameko.context': 'data',
            'nameko.call_id_stack': ['service-name.method.0'],
        }

        value = {'foo': Mock()}

        service_rpc = ServiceRpc(
            "service-name", **{'headers': value}
        ).bind(mock_container, "service_rpc")

        service_rpc.reply_listener.setup()
        service_rpc.setup()

        service_rpc = service_rpc.get_dependency(worker_ctx)

        def merge_dicts(base, *updates):
            merged = base.copy()
            [merged.update(update) for update in updates]
            return merged

        service_rpc.method.call_async()
        assert producer.publish.call_args[1]['headers'] == merge_dicts(
            nameko_headers, value
        )

    @pytest.mark.usefixtures("memory_rabbit_config")
    @patch('nameko.rpc.uuid')
    def test_restricted_parameters(self, patch_uuid, mock_container, producer):
        """ Verify that providing routing parameters at instantiation
        time has no effect.
        """
        mock_container.shared_extensions = {}
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.container = mock_container
        worker_ctx.context_data = {}

        patch_uuid.uuid4.side_effect = ["uuid1", "uuid2"]

        restricted_params = (
            'exchange', 'routing_key', 'mandatory',
            'correlation_id', 'reply_to'
        )

        service_rpc = ServiceRpc(
            "service", **{param: Mock() for param in restricted_params}
        ).bind(mock_container, "service_rpc")

        service_rpc.reply_listener.setup()
        service_rpc.setup()

        service_rpc = service_rpc.get_dependency(worker_ctx)

        service_rpc.method.call_async()
        publish_params = producer.publish.call_args[1]

        assert publish_params['exchange'].name == "nameko-rpc"
        assert publish_params['routing_key'] == 'service.method'
        assert publish_params['mandatory'] is True
        assert publish_params['reply_to'] == "uuid1"
        assert publish_params['correlation_id'] == "uuid2"


@pytest.mark.usefixtures("rabbit_config")
@config.patch({MAX_WORKERS_CONFIG_KEY: 1})
def test_prefetch_throughput(container_factory):
    """Make sure even max_workers=1 can consumer faster than 1 msg/second

    Regression test for https://github.com/nameko/nameko/issues/417
    """

    class Service(object):
        name = "service"

        @rpc
        def method(self):
            pass

    container = container_factory(Service)
    container.start()

    replies = []
    with ServiceRpcClient("service") as client:
        for _ in range(5):
            replies.append(client.method.call_async())

        with eventlet.Timeout(1):
            [reply.result() for reply in replies]


@pytest.mark.usefixtures("rabbit_config")
def test_stop_with_active_worker(container_factory, queue_info):
    """ Test behaviour when we stop a container with an active worker.

    Expect the consumer to stop and the message be requeued, but the container
    to continue running the active worker until it completes.

    This is not desirable behaviour but it is consistent with the old
    implementation. It would be better to stop the consumer but keep the
    channel alive until the worker has completed and the message can be
    ack'd, but we can't do that with kombu or without per-entrypoint worker
    pools.
    """

    block = Event()

    class Service(object):
        name = "service"

        @rpc
        def method(self):
            block.wait()

    container = container_factory(Service)
    container.start()

    with ServiceRpcClient("service") as service_rpc:
        service_rpc.method.call_async()

    gt = eventlet.spawn(container.stop)

    @retry
    def consumer_removed():
        info = queue_info('rpc-service')
        assert info.consumer_count == 0
        assert info.message_count == 1

    consumer_removed()

    assert not gt.dead
    block.send(True)

    gt.wait()
    assert gt.dead


class TestSSL(object):

    @pytest.fixture(params=[True, False])
    def rabbit_ssl_options(self, request, rabbit_ssl_options):
        verify_certs = request.param
        if verify_certs is False:
            # remove certificate paths from config
            options = True
        else:
            options = rabbit_ssl_options
        return options

    @pytest.mark.usefixtures("rabbit_ssl_config")
    def test_rpc_entrypoint_over_ssl(self, container_factory, rabbit_uri):

        class Service(object):
            name = "service"

            @rpc
            def echo(self, *args, **kwargs):
                return args, kwargs

        container = container_factory(Service)
        container.start()

        with ServiceRpcClient("service", uri=rabbit_uri, ssl=False) as client:
            assert client.echo("a", "b", foo="bar") == [
                ['a', 'b'], {'foo': 'bar'}
            ]

    @pytest.mark.usefixtures("rabbit_config")
    def test_rpc_client_over_ssl(
        self, container_factory, rabbit_ssl_uri, rabbit_ssl_options
    ):
        class Service(object):
            name = "service"

            delegate_rpc = ServiceRpc(
                'delegate',
                uri=rabbit_ssl_uri,
                ssl=rabbit_ssl_options
            )

            @dummy
            def echo(self, *args, **kwargs):
                return self.delegate_rpc.echo(*args, **kwargs)

        class Delegate(object):
            name = "delegate"

            @rpc
            def echo(self, *args, **kwargs):
                return args, kwargs

        container = container_factory(Service)
        container.start()

        delegate = container_factory(Delegate)
        delegate.start()

        with entrypoint_hook(container, 'echo') as echo:
            assert echo("a", "b", foo="bar") == [
                ['a', 'b'], {'foo': 'bar'}
            ]


class TestEntrypointArguments:

    @pytest.mark.usefixtures("rabbit_config")
    def test_expected_exceptions_and_sensitive_arguments(
        self, container_factory, rabbit_config
    ):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @rpc(expected_exceptions=Boom, sensitive_arguments=["arg"])
            def method(self, arg):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, Rpc)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["arg"]



import re
import socket
from collections import namedtuple
from functools import partial
from logging import getLogger

import eventlet
from eventlet import wsgi
from eventlet.support import get_errno
from eventlet.wsgi import BROKEN_SOCK, BaseHTTPServer, HttpProtocol
from werkzeug.exceptions import HTTPException
from werkzeug.routing import Map
from werkzeug.wrappers import Request

from nameko import config
from nameko.constants import WEB_SERVER_CONFIG_KEY
from nameko.exceptions import ConfigurationError
from nameko.extensions import ProviderCollector, SharedExtension


try:
    STATE_IDLE = wsgi.STATE_IDLE
except AttributeError:  # pragma: no cover
    STATE_IDLE = None


BindAddress = namedtuple("BindAddress", ['address', 'port'])


def parse_address(address_string):
    address_re = re.compile(r'^((?P<address>[^:]+):)?(?P<port>\d+)$')
    match = address_re.match(address_string)
    if match is None:
        raise ConfigurationError(
            'Misconfigured bind address `{}`. '
            'Should be `[address:]port`'.format(address_string)
        )
    address = match.group('address') or ''
    port = int(match.group('port'))
    return BindAddress(address, port)


class HttpOnlyProtocol(HttpProtocol):
    # identical to HttpProtocol.finish, except greenio.shutdown_safe
    # is removed. it's only needed to ssl sockets which we don't support
    # this is a workaround until
    # https://bitbucket.org/eventlet/eventlet/pull-request/42
    # or something like it lands
    def finish(self):
        try:
            # patched in depending on python version; confuses pylint
            # pylint: disable=E1101
            BaseHTTPServer.BaseHTTPRequestHandler.finish(self)
        except socket.error as e:
            # Broken pipe, connection reset by peer
            if get_errno(e) not in BROKEN_SOCK:
                raise
        self.connection.close()


class WebServer(ProviderCollector, SharedExtension):
    """A SharedExtension that wraps a WSGI interface for processing HTTP
    requests.

    WebServer can be subclassed to add additional WSGI functionality through
    overriding the `get_wsgi_server` and `get_wsgi_app` methods.
    """

    def __init__(self):
        super(WebServer, self).__init__()
        self._gt = None
        self._sock = None
        self._serv = None
        self._starting = False
        self._is_accepting = True

    @property
    def bind_addr(self):
        address_str = config.get(
            WEB_SERVER_CONFIG_KEY, '0.0.0.0:8000')
        return parse_address(address_str)

    def run(self):
        while self._is_accepting:
            sock, addr = self._sock.accept()
            sock.settimeout(self._serv.socket_timeout)
            self.container.spawn_managed_thread(
                partial(self.process_request, sock, addr)
            )

    def process_request(self, sock, address):
        try:
            if STATE_IDLE:  # pragma: no cover
                # eventlet >= 0.22
                # see https://github.com/eventlet/eventlet/issues/420
                self._serv.process_request([address, sock, STATE_IDLE])
            else:  # pragma: no cover
                self._serv.process_request((sock, address))

        except OSError as exc:
            # OSError("raw readinto() returned invalid length")
            # can be raised when a client disconnects very early as a result
            # of an eventlet bug: https://github.com/eventlet/eventlet/pull/353
            # See https://github.com/onefinestay/nameko/issues/368
            if "raw readinto() returned invalid length" in str(exc):
                return
            raise

    def start(self):
        if not self._starting:
            self._starting = True
            self._sock = eventlet.listen(self.bind_addr)
            # work around https://github.com/celery/kombu/issues/838
            self._sock.settimeout(None)
            self._serv = self.get_wsgi_server(self._sock, self.get_wsgi_app())
            self._gt = self.container.spawn_managed_thread(self.run)

    def get_wsgi_app(self):
        """Get the WSGI application used to process requests.

        This method can be overridden to apply WSGI middleware or replace
        the WSGI application all together.
        """
        return WsgiApp(self)

    def get_wsgi_server(
        self, sock, wsgi_app, protocol=HttpOnlyProtocol, debug=False
    ):
        """Get the WSGI server used to process requests."""
        return wsgi.Server(
            sock,
            sock.getsockname(),
            wsgi_app,
            protocol=protocol,
            debug=debug,
            log=getLogger(__name__)
        )

    def stop(self):
        self._is_accepting = False
        self._gt.kill()
        self._sock.close()
        super(WebServer, self).stop()

    def make_url_map(self):
        url_map = Map()
        for provider in self._providers:
            rule = provider.get_url_rule()
            rule.endpoint = provider
            url_map.add(rule)
        return url_map

    def context_data_from_headers(self, request):
        return {}


class WsgiApp(object):

    def __init__(self, server):
        self.server = server
        self.url_map = server.make_url_map()

    def __call__(self, environ, start_response):
        # set to shallow mode so that nobody can read request data in
        # until unset.  This allows the connection to be upgraded into
        # a bidirectional websocket connection later if we need in all
        # cases.  The regular request handling code unsets this flag
        # automatically.
        #
        # If we would not do this and some code would access the form data
        # before that point, we might deadlock outselves because the
        # browser is no longer reading from our socket at this point in
        # time.
        request = Request(environ, shallow=True)
        adapter = self.url_map.bind_to_environ(environ)
        try:
            provider, values = adapter.match()
            request.path_values = values
            rv = provider.handle_request(request)
        except HTTPException as exc:
            rv = exc
        return rv(environ, start_response)



from __future__ import absolute_import

import inspect
import types
import weakref
from functools import partial
from logging import getLogger

from eventlet.event import Event

from nameko.exceptions import IncorrectSignature


_log = getLogger(__name__)


ENTRYPOINT_EXTENSIONS_ATTR = 'nameko_entrypoints'


class Extension(object):
    """ Note that Extension.__init__ is called during :meth:`bind` as
    well as at instantiation time, so avoid side-effects in this method.
    Use :meth:`setup` instead.

    Furthermore, :meth:`bind` and :func:`iter_extensions` use introspection
    to find any sub-extensions that an extension may declare. Any descriptors
    on the extension should expect to be called during introspection, which
    happens between `ServiceContainer.__init__` and `ServiceContainer.setup`.

    :attr:`Extension.container` gives access to the
    :class:`~nameko.containers.ServiceContainer` instance to
    which the Extension is bound, otherwise `None`.
    """

    __params = None
    container = None

    def __new__(cls, *args, **kwargs):
        inst = super(Extension, cls).__new__(cls)
        inst.__params = (args, kwargs)
        return inst

    def setup(self):
        """ Called on bound Extensions before the container starts.

        Extensions should do any required initialisation here.
        """

    def start(self):
        """ Called on bound Extensions when the container has successfully
        started.

        This is only called after all other Extensions have successfully
        returned from :meth:`Extension.setup`. If the Extension reacts
        to external events, it should now start acting upon them.
        """

    def stop(self):
        """ Called when the service container begins to shut down.

        Extensions should do any graceful shutdown here.
        """

    def kill(self):
        """ Called to stop this extension without grace.

        Extensions should urgently shut down here. This means
        stopping as soon as possible by omitting cleanup.
        This may be distinct from ``stop()`` for certain dependencies.

        Extensions should not raise during kill, since the container
        is already dying. Instead they should log what is appropriate and
        swallow the exception to allow the container kill to continue.
        """

    def bind(self, container):
        """ Get an instance of this Extension to bind to `container`.
        """

        def clone(prototype):
            if prototype.is_bound():
                raise RuntimeError('Cannot `bind` a bound extension.')

            cls = type(prototype)
            args, kwargs = prototype.__params
            instance = cls(*args, **kwargs)
            # instance.container must be a weakref to avoid a strong reference
            # from value to key in the `shared_extensions` weakkey dict
            # see test_extension_sharing.py: test_weakref
            instance.container = weakref.proxy(container)
            return instance

        instance = clone(self)

        # recurse over sub-extensions
        for name, ext in inspect.getmembers(self, is_extension):
            setattr(instance, name, ext.bind(container))
        return instance

    def is_bound(self):
        return self.container is not None

    def __repr__(self):
        if not self.is_bound():
            return '<{} [unbound] at 0x{:x}>'.format(
                type(self).__name__, id(self))

        return '<{} at 0x{:x}>'.format(
            type(self).__name__, id(self))


class SharedExtension(Extension):

    @property
    def sharing_key(self):
        return type(self)

    def bind(self, container):
        """ Bind implementation that supports sharing.
        """
        # if there's already a matching bound instance, return that
        shared = container.shared_extensions.get(self.sharing_key)
        if shared:
            return shared

        instance = super(SharedExtension, self).bind(container)

        # save the new instance
        container.shared_extensions[self.sharing_key] = instance

        return instance


class DependencyProvider(Extension):

    attr_name = None

    def bind(self, container, attr_name):
        """ Get an instance of this Dependency to bind to `container` with
        `attr_name`.
        """
        instance = super(DependencyProvider, self).bind(container)
        instance.attr_name = attr_name
        self.attr_name = attr_name
        return instance

    def get_dependency(self, worker_ctx):
        """ Called before worker execution. A DependencyProvider should return
        an object to be injected into the worker instance by the container.
        """

    def worker_result(self, worker_ctx, result=None, exc_info=None):
        """ Called with the result of a service worker execution.

        Dependencies that need to process the result should do it here.
        This method is called for all `Dependency` instances on completion
        of any worker.

        Example: a database session dependency may flush the transaction

        :Parameters:
            worker_ctx : WorkerContext
                See ``nameko.containers.ServiceContainer.spawn_worker``
        """

    def worker_setup(self, worker_ctx):
        """ Called before a service worker executes a task.

        Dependencies should do any pre-processing here, raising exceptions
        in the event of failure.

        Example: ...

        :Parameters:
            worker_ctx : WorkerContext
                See ``nameko.containers.ServiceContainer.spawn_worker``
        """

    def worker_teardown(self, worker_ctx):
        """ Called after a service worker has executed a task.

        Dependencies should do any post-processing here, raising
        exceptions in the event of failure.

        Example: a database session dependency may commit the session

        :Parameters:
            worker_ctx : WorkerContext
                See ``nameko.containers.ServiceContainer.spawn_worker``
        """

    def __repr__(self):
        if not self.is_bound():
            return '<{} [unbound] at 0x{:x}>'.format(
                type(self).__name__, id(self))

        service_name = self.container.service_name
        return '<{} [{}.{}] at 0x{:x}>'.format(
            type(self).__name__, service_name, self.attr_name, id(self))


class ProviderCollector(object):
    def __init__(self, *args, **kwargs):
        self._providers = set()
        self._providers_registered = False
        self._last_provider_unregistered = Event()
        super(ProviderCollector, self).__init__(*args, **kwargs)

    def register_provider(self, provider):
        self._providers_registered = True
        _log.debug('registering provider %s for %s', provider, self)
        self._providers.add(provider)

    def unregister_provider(self, provider):
        providers = self._providers
        if provider not in self._providers:
            return

        _log.debug('unregistering provider %s for %s', provider, self)

        providers.remove(provider)
        if len(providers) == 0:
            _log.debug('last provider unregistered for %s', self)
            self._last_provider_unregistered.send()

    def wait_for_providers(self):
        """ Wait for any providers registered with the collector to have
        unregistered.

        Returns immediately if no providers were ever registered.
        """
        if self._providers_registered:
            _log.debug('waiting for providers to unregister %s', self)
            self._last_provider_unregistered.wait()
            _log.debug('all providers unregistered %s', self)

    def stop(self):
        """ Default `:meth:Extension.stop()` implementation for
        subclasses using `ProviderCollector` as a mixin.
        """
        self.wait_for_providers()


def register_entrypoint(fn, entrypoint):
    descriptors = getattr(fn, ENTRYPOINT_EXTENSIONS_ATTR, None)

    if descriptors is None:
        descriptors = set()
        setattr(fn, ENTRYPOINT_EXTENSIONS_ATTR, descriptors)

    descriptors.add(entrypoint)


class Entrypoint(Extension):

    method_name = None

    def __init__(
        self, expected_exceptions=(), sensitive_arguments=(), **kwargs
    ):
        """
        :Parameters:
            expected_exceptions : exception class or tuple of exception classes
                Specify exceptions that may be caused by the caller (e.g. by
                providing bad arguments). Saved on the entrypoint instance as
                ``entrypoint.expected_exceptions`` for later inspection by
                other extensions, for example a monitoring system.
            sensitive_arguments : string or tuple of strings
                Mark an argument or part of an argument as sensitive. Saved on
                the entrypoint instance as ``entrypoint.sensitive_arguments``
                for later inspection by other extensions, for example a
                logging system.

                :seealso: :func:`nameko.utils.get_redacted_args`
        """
        self.expected_exceptions = expected_exceptions
        self.sensitive_arguments = sensitive_arguments
        super(Entrypoint, self).__init__(**kwargs)

    def bind(self, container, method_name):
        """ Get an instance of this Entrypoint to bind to `container` with
        `method_name`.
        """
        instance = super(Entrypoint, self).bind(container)
        instance.method_name = method_name
        return instance

    def check_signature(self, args, kwargs):
        service_cls = self.container.service_cls
        fn = getattr(service_cls, self.method_name)
        try:
            service_instance = None  # fn is unbound
            inspect.getcallargs(fn, service_instance, *args, **kwargs)
        except TypeError as exc:
            raise IncorrectSignature(str(exc))

    @classmethod
    def decorator(cls, *args, **kwargs):

        def registering_decorator(fn, args, kwargs):
            instance = cls(*args, **kwargs)
            register_entrypoint(fn, instance)
            return fn

        if len(args) == 1 and isinstance(args[0], types.FunctionType):
            # usage without arguments to the decorator:
            # @foobar
            # def spam():
            #     pass
            return registering_decorator(args[0], args=(), kwargs={})
        else:
            # usage with arguments to the decorator:
            # @foobar('shrub', ...)
            # def spam():
            #     pass
            return partial(registering_decorator, args=args, kwargs=kwargs)

    def __repr__(self):
        if not self.is_bound():
            return '<{} [unbound] at 0x{:x}>'.format(
                type(self).__name__, id(self))

        service_name = self.container.service_name
        return '<{} [{}.{}] at 0x{:x}>'.format(
            type(self).__name__, service_name, self.method_name, id(self))


def is_extension(obj):
    return isinstance(obj, Extension)


def is_dependency(obj):
    return isinstance(obj, DependencyProvider)


def is_entrypoint(obj):
    return isinstance(obj, Entrypoint)


def iter_extensions(extension):
    """ Depth-first iterator over sub-extensions on `extension`.
    """
    for _, ext in inspect.getmembers(extension, is_extension):
        for item in iter_extensions(ext):
            yield item
        yield ext



import sys
import time
from threading import Thread

import pytest
from mock import ANY, Mock, call, patch

from nameko.testing.waiting import WaitResult, wait_for_call


@pytest.yield_fixture
def forever():
    value = [True]
    yield value
    value.pop()


class TestPatchWaitUseCases(object):

    def test_wait_for_specific_result(self, forever):

        class Counter(object):
            value = 0

            def count(self):
                self.value += 1
                return self.value

        counter = Counter()

        def count_forever():
            while forever:
                counter.count()
                time.sleep(0)

        def cb(args, kwargs, res, exc_info):
            return res == 10

        with wait_for_call(counter, 'count', callback=cb) as result:
            Thread(target=count_forever).start()

        assert result.get() == 10

    def test_wait_until_called_with_argument(self, forever):

        class CounterWithSet(object):

            value = 0

            def set(self, value):
                self.value = value
                return self.value

        counter = CounterWithSet()

        def increment_forever_via_set():
            while forever:
                counter.set(counter.value + 1)
                time.sleep(0)

        def cb(args, kwargs, res, exc_info):
            return args == (10,)

        with wait_for_call(counter, 'set', callback=cb) as result:
            Thread(target=increment_forever_via_set).start()

        assert result.get() == 10

    def test_wait_until_raises(self, forever):

        class LimitExceeded(Exception):
            pass

        class CounterWithLimit(object):
            def __init__(self, limit):
                self.value = 0
                self.limit = limit

            def count(self):
                self.value += 1
                if self.value >= self.limit:
                    raise LimitExceeded(self.limit)
                return self.value

        limit = 10
        counter = CounterWithLimit(limit)

        def count_forever():
            while forever:
                counter.count()
                time.sleep(0)

        def cb(args, kwargs, res, exc_info):
            return exc_info is not None

        with wait_for_call(counter, 'count', callback=cb) as result:
            Thread(target=count_forever).start()

        with pytest.raises(LimitExceeded):
            result.get()

    def test_wait_until_stops_raising(self, forever):

        class ThresholdNotReached(Exception):
            pass

        class CounterWithThreshold(object):

            def __init__(self, threshold):
                self.value = 0
                self.threshold = threshold

            def count(self):
                self.value += 1
                if self.value < self.threshold:
                    raise ThresholdNotReached(self.threshold)
                return self.value

        threshold = 10
        counter = CounterWithThreshold(threshold)

        def count_forever():
            while forever:
                try:
                    counter.count()
                except ThresholdNotReached:
                    pass
                time.sleep(0)

        def cb(args, kwargs, res, exc_info):
            return exc_info is None

        with wait_for_call(counter, 'count', callback=cb) as result:
            Thread(target=count_forever).start()

        assert result.get() == threshold


class TestPatchWait(object):

    def test_direct(self):

        class Echo(object):

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        with wait_for_call(echo, 'upper'):
            res = echo.upper(arg)
            assert res == "HELLO"

    def test_indirect(self):

        class Echo(object):

            def proxy(self, arg):
                return self.upper(arg)

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        with wait_for_call(echo, 'upper'):
            assert echo.proxy(arg) == "HELLO"

    def test_patch_class(self):

        class Echo(object):

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        with wait_for_call(Echo, 'upper'):
            res = echo.upper(arg)
            assert res == "HELLO"

    def test_result(self):

        class Echo(object):

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        with wait_for_call(echo, 'upper') as result:
            res = echo.upper(arg)

        assert result.get() == res

    def test_result_not_ready(self):

        class Echo(object):

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        with wait_for_call(echo, 'upper') as result:
            with pytest.raises(result.NotReady):
                result.get()
            res = echo.upper(arg)

        assert result.get() == res

    def test_result_is_none(self):

        class Echo(object):

            def nothing(self):
                return None

        echo = Echo()

        with wait_for_call(echo, 'nothing') as result:
            res = echo.nothing()

        assert res is None
        assert result.get() is None
        assert result.has_result is True

    def test_wrapped_method_raises(self):

        class EchoException(Exception):
            pass

        class Echo(object):

            def error(self):
                raise EchoException("error!")

        echo = Echo()

        with wait_for_call(echo, 'error'):
            with pytest.raises(EchoException):
                echo.error()

    def test_result_get_raises(self):

        class EchoException(Exception):
            pass

        class Echo(object):

            def error(self):
                raise EchoException("error!")

        echo = Echo()

        with wait_for_call(echo, 'error') as result:
            with pytest.raises(EchoException):
                echo.error()

            with pytest.raises(EchoException):
                result.get()

    def test_callback(self):

        class Echo(object):

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        callback = Mock()
        callback.return_value = True

        with wait_for_call(echo, 'upper', callback):
            res = echo.upper(arg)
            assert res == "HELLO"

        assert callback.called
        assert callback.call_args_list == [call((arg,), {}, res, None)]

    def test_callback_multiple_calls(self):

        class Echo(object):

            count = 0

            def upper(self, arg):
                self.count += 1
                return "{}-{}".format(arg.upper(), self.count)

        echo = Echo()
        arg = "hello"

        callback = Mock()
        callback.side_effect = [False, True]

        with wait_for_call(echo, 'upper', callback):
            res1 = echo.upper(arg)
            assert res1 == "HELLO-1"
            res2 = echo.upper(arg)
            assert res2 == "HELLO-2"

        assert callback.called
        assert callback.call_args_list == [
            call((arg,), {}, res1, None),
            call((arg,), {}, res2, None),
        ]

    def test_callback_with_exception(self):

        class EchoException(Exception):
            pass

        class Echo(object):

            def error(self):
                raise exc

        echo = Echo()
        exc = EchoException("error!")

        callback = Mock()
        callback.return_value = True

        with wait_for_call(echo, 'error', callback):
            with pytest.raises(EchoException):
                echo.error()

        assert callback.called
        assert callback.call_args_list == [
            call((), {}, None, (EchoException, exc, ANY))
        ]

    def test_callback_with_exception_multiple_calls(self):

        class EchoException(Exception):
            pass

        class Echo(object):

            def error(self):
                raise exc

        echo = Echo()
        exc = EchoException("error!")

        callback = Mock()
        callback.side_effect = [False, True]

        with wait_for_call(echo, 'error', callback):
            with pytest.raises(EchoException):
                echo.error()
            with pytest.raises(EchoException):
                echo.error()

        assert callback.called
        assert callback.call_args_list == [
            call((), {}, None, (EchoException, exc, ANY)),
            call((), {}, None, (EchoException, exc, ANY))
        ]

    def test_with_new_thread(self):

        class Echo(object):

            def proxy(self, arg):
                Thread(target=self.upper, args=(arg,)).start()

            def upper(self, arg):
                return arg.upper()

        echo = Echo()
        arg = "hello"

        callback = Mock()
        callback.return_value = True

        with wait_for_call(echo, 'upper', callback):
            res = echo.proxy(arg)
            assert res is None

        assert callback.called
        assert callback.call_args_list == [call((arg,), {}, "HELLO", None)]

    def test_target_as_mock(self):

        class Klass(object):

            def __init__(self):
                self.attr = "value"

            def method(self):
                return self.attr.upper()

        instance = Klass()

        with patch.object(instance, 'attr') as patched_attr:

            with wait_for_call(patched_attr, 'upper'):
                instance.method()

            assert patched_attr.upper.called
            assert instance.attr.upper.called


class TestWaitResult(object):

    class CustomError(Exception):
        pass

    @pytest.fixture
    def exc_info(self):
        try:
            raise self.CustomError("whoops")
        except:
            exc_info = sys.exc_info()
        return exc_info

    def test_has_result(self):
        result = WaitResult()
        assert result.has_result is False
        result.send("ok", None)
        assert result.has_result is True

    def test_has_exception(self, exc_info):
        result = WaitResult()
        assert result.has_result is False
        result.send(None, exc_info)
        assert result.has_result is True

    def test_send_multiple_times(self):
        result = WaitResult()
        result.send(1, None)
        result.send(2, None)
        assert result.get() == 1

    def test_get_result_multiple_times(self):
        result = WaitResult()
        result.send(1, None)
        assert result.get() == 1
        assert result.get() == 1

    def test_get_raises(self, exc_info):

        result = WaitResult()
        result.send(1, exc_info)
        with pytest.raises(self.CustomError):
            result.get()



# coding: utf-8

import eventlet
import pytest

from nameko.extensions import ProviderCollector


def test_provider_collector():
    collector = ProviderCollector()

    provider1 = object()
    provider2 = object()
    collector.register_provider(provider1)
    collector.register_provider(provider2)

    assert provider1 in collector._providers
    assert provider2 in collector._providers

    collector.unregister_provider(provider1)
    assert provider1 not in collector._providers

    # unregister missing provider is a no-op
    collector.unregister_provider(provider1)

    # stop() should block while a provider remains
    with pytest.raises(eventlet.Timeout):
        with eventlet.Timeout(0):
            collector.stop()

    # stop() will return when the final provider is unregistered
    with eventlet.Timeout(0):
        collector.unregister_provider(provider2)
        collector.stop()



import pytest

from nameko import config


@pytest.mark.usefixtures("empty_config")
def test_setup():

    assert config == {}

    config.setup({"HEARTBEAT": 10})

    assert config == {"HEARTBEAT": 10}

    config.setup({})

    assert config == {}


@pytest.mark.usefixtures("empty_config")
def test_patch_full_context_manager():

    config.setup({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    with config.patch({"HEARTBEAT": 40}, clear=True):
        assert config == {"HEARTBEAT": 40}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}


@pytest.mark.usefixtures("empty_config")
def test_patch_full_nesting_context_managers():

    with config.patch({"HEARTBEAT": 10}):
        assert config == {"HEARTBEAT": 10}

    assert config == {}

    with config.patch({"HEARTBEAT": 40}, clear=True):
        assert config == {"HEARTBEAT": 40}
        with config.patch({"HEARTBEAT": 10}, clear=True):
            assert config == {"HEARTBEAT": 10}
        assert config == {"HEARTBEAT": 40}

    assert config == {}


@pytest.mark.usefixtures("empty_config")
def test_patch_full_decorator():

    config.setup({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    @config.patch({"HEARTBEAT": 40}, clear=True)
    def test():
        assert config == {"HEARTBEAT": 40}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}

    test()

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}


@pytest.mark.usefixtures("empty_config")
def test_patch_context_manager():

    config.setup({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    with config.patch({"HEARTBEAT": 40}):
        assert config == {"HEARTBEAT": 40, "ACCEPT": "YAML"}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}


@pytest.mark.usefixtures("empty_config")
def test_patch_decorator():

    config.setup({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    @config.patch({"HEARTBEAT": 40})
    def test():
        assert config == {"HEARTBEAT": 40, "ACCEPT": "YAML"}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}

    test()

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}


@pytest.mark.usefixtures("empty_config")
def test_patch_decorator_deals_with_failing_function_execution():

    config.setup({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    exception = Exception('Boom!')

    @config.patch({"HEARTBEAT": 40})
    def test():
        assert config == {"HEARTBEAT": 40, "ACCEPT": "YAML"}
        raise exception

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}

    with pytest.raises(Exception) as exc_info:
        test()

    assert exc_info.value == exception

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}



""" Service unit testing best practice.
"""

from nameko.rpc import ServiceRpc, rpc
from nameko.testing.services import worker_factory


class ConversionService(object):
    """ Service under test
    """
    name = "conversions"

    maths_rpc = ServiceRpc("maths")

    @rpc
    def inches_to_cm(self, inches):
        return self.maths_rpc.multiply(inches, 2.54)

    @rpc
    def cms_to_inches(self, cms):
        return self.maths_rpc.divide(cms, 2.54)


def test_conversion_service():
    # create worker with mock dependencies
    service = worker_factory(ConversionService)

    # add side effects to the mock rpc dependency on the "maths" service
    service.maths_rpc.multiply.side_effect = lambda x, y: x * y
    service.maths_rpc.divide.side_effect = lambda x, y: x / y

    # test inches_to_cm business logic
    assert service.inches_to_cm(300) == 762
    service.maths_rpc.multiply.assert_called_once_with(300, 2.54)

    # test cms_to_inches business logic
    assert service.cms_to_inches(762) == 300
    service.maths_rpc.divide.assert_called_once_with(762, 2.54)



import sys

import eventlet
from eventlet.queue import LightQueue


def fail_fast_imap(pool, call, items):
    """ Run a function against each item in a given list, yielding each
    function result in turn, where the function call is handled in a
    :class:`~eventlet.greenthread.GreenThread` spawned by the provided pool.

    If any function raises an exception, all other ongoing threads are killed,
    and the exception is raised to the caller.

    This function is similar to :meth:`~eventlet.greenpool.GreenPool.imap`.

    :param pool: Pool to spawn function threads from
    :type pool: eventlet.greenpool.GreenPool
    :param call: Function call to make, expecting to receive an item from the
        given list
    """
    result_queue = LightQueue(maxsize=len(items))
    spawned_threads = set()

    def handle_result(finished_thread):
        try:
            thread_result = finished_thread.wait()
            spawned_threads.remove(finished_thread)
            result_queue.put((thread_result, None))
        except Exception:
            spawned_threads.remove(finished_thread)
            result_queue.put((None, sys.exc_info()))

    for item in items:
        gt = pool.spawn(call, item)
        spawned_threads.add(gt)
        gt.link(handle_result)

    while spawned_threads:
        result, exc_info = result_queue.get()
        if exc_info is not None:
            # Kill all other ongoing threads
            for ongoing_thread in spawned_threads:
                ongoing_thread.kill()
            # simply raising here (even raising a full exc_info) isn't
            # sufficient to preserve the original stack trace.
            # greenlet.throw() achieves this.
            eventlet.getcurrent().throw(*exc_info)
        yield result


class SpawningProxy(object):
    def __init__(self, items, abort_on_error=False):
        """ Wraps an iterable set of items such that a call on the returned
        SpawningProxy instance will spawn a call in a
        :class:`~eventlet.greenthread.GreenThread` for each item.

        Returns when every spawned thread has completed.

        :param items: Iterable item set to process
        :param abort_on_error: If True, any exceptions raised on an individual
            item call will cause all peer item call threads to be killed, and
            for the exception to be propagated to the caller immediately.
        """
        self._items = items
        self.abort_on_error = abort_on_error

    def __getattr__(self, name):

        def spawning_method(*args, **kwargs):
            items = self._items
            if items:
                pool = eventlet.GreenPool(len(items))

                def call(item):
                    return getattr(item, name)(*args, **kwargs)

                if self.abort_on_error:
                    return list(fail_fast_imap(pool, call, self._items))
                else:
                    return list(pool.imap(call, self._items))
        return spawning_method


class SpawningSet(set):
    """ A set with an ``.all`` property that will spawn a method call on each
    item in the set into its own (parallel) greenthread.
    """
    @property
    def all(self):
        return SpawningProxy(self)



from __future__ import absolute_import

import itertools
import time
from logging import getLogger

from eventlet import Timeout
from eventlet.event import Event

from nameko.extensions import Entrypoint


_log = getLogger(__name__)


class Timer(Entrypoint):
    def __init__(self, interval, eager=False, **kwargs):
        """
        Timer entrypoint. Fires every `interval` seconds or as soon as
        the previous worker completes if that took longer.

        The default behaviour is to wait `interval` seconds
        before firing for the first time. If you want the entrypoint
        to fire as soon as the service starts, pass `eager=True`.

        Example::

            timer = Timer.decorator

            class Service(object):
                name = "service"

                @timer(interval=5)
                def tick(self):
                    pass

        """
        self.interval = interval
        self.eager = eager
        self.should_stop = Event()
        self.worker_complete = Event()
        self.gt = None
        super(Timer, self).__init__(**kwargs)

    def start(self):
        _log.debug('starting %s', self)
        self.gt = self.container.spawn_managed_thread(self._run)

    def stop(self):
        _log.debug('stopping %s', self)
        self.should_stop.send(True)
        self.gt.wait()

    def kill(self):
        _log.debug('killing %s', self)
        self.gt.kill()

    def _run(self):
        """ Runs the interval loop. """

        def get_next_interval():
            start_time = time.time()
            start = 0 if self.eager else 1
            for count in itertools.count(start=start):
                yield max(start_time + count * self.interval - time.time(), 0)
        interval = get_next_interval()
        sleep_time = next(interval)
        while True:
            # sleep for `sleep_time`, unless `should_stop` fires, in which
            # case we leave the while loop and stop entirely
            with Timeout(sleep_time, exception=False):
                self.should_stop.wait()
                break

            self.handle_timer_tick()

            self.worker_complete.wait()
            self.worker_complete.reset()

            sleep_time = next(interval)

    def handle_timer_tick(self):
        args = ()
        kwargs = {}

        # Note that we don't catch ContainerBeingKilled here. If that's raised,
        # there is nothing for us to do anyway. The exception bubbles, and is
        # caught by :meth:`Container._handle_thread_exited`, though the
        # triggered `kill` is a no-op, since the container is already
        # `_being_killed`.
        self.container.spawn_worker(
            self, args, kwargs, handle_result=self.handle_result)

    def handle_result(self, worker_ctx, result, exc_info):
        self.worker_complete.send()
        return result, exc_info


timer = Timer.decorator



from __future__ import absolute_import

import pytest
import six

from nameko import config


# all imports are inline to make sure they happen after eventlet.monkey_patch
# which is called in pytest_load_initial_conftests
# (calling monkey_patch at import time breaks the pytest capturemanager - see
#  https://github.com/eventlet/eventlet/pull/239)


def parse_config_option(text):
    import yaml
    if '=' in text:
        key, value = text.strip().split('=', 1)
        return key, yaml.unsafe_load(value)
    else:
        return text, True


def pytest_addoption(parser):
    import ssl
    parser.addoption(
        '--blocking-detection',
        action='store_true',
        dest='blocking_detection',
        default=False,
        help='turn on eventlet hub blocking detection')

    parser.addoption(
        "--amqp-uri", "--rabbit-amqp-uri",
        action="store",
        dest='RABBIT_AMQP_URI',
        default='pyamqp://guest:guest@localhost:5672/',
        help=(
            "URI for the RabbitMQ broker. Any specified virtual host will be "
            "ignored because tests run in their own isolated vhost."
        ))

    parser.addoption(
        "--rabbit-api-uri", "--rabbit-ctl-uri",
        action="store",
        dest='RABBIT_API_URI',
        default='http://guest:guest@localhost:15672',
        help=("URI for RabbitMQ management interface.")
    )

    parser.addoption(
        '--amqp-ssl-port',
        action='store',
        dest='AMQP_SSL_PORT',
        default=5671,
        help='Port number for SSL connection')

    parser.addoption(
        '--amqp-ssl-option',
        type=parse_config_option,
        action='append',
        dest='AMQP_SSL_OPTIONS',
        metavar='KEY=VALUE',
        default=[
            ('ca_certs', 'certs/cacert.pem'),
            ('certfile', 'certs/clientcert.pem'),
            ('keyfile', 'certs/clientkey.pem'),
            ('cert_reqs', ssl.CERT_REQUIRED)
        ],
        help=(
            'SSL connection options for passing to ssl.wrap_socket.'
            'Multiple options may be given. Values are parsed as YAML, '
            'hence the following example is valid: \n'
            '--amqp-ssl-option certfile=clientcert.pem '
            '--amqp-ssl-option ssl_version=!!python/name:ssl.PROTOCOL_TLSv1_2'
        )
    )


def pytest_load_initial_conftests():
    # make sure we monkey_patch before local conftests
    import eventlet
    eventlet.monkey_patch()


def pytest_configure(config):
    if config.option.blocking_detection:  # pragma: no cover
        from eventlet import debug
        debug.hub_blocking_detection(True)


@pytest.yield_fixture
def empty_config():
    with config.patch({}, clear=True):
        yield


@pytest.yield_fixture
def mock_container(request):
    from mock import create_autospec
    from nameko.constants import SERIALIZER_CONFIG_KEY, DEFAULT_SERIALIZER
    from nameko.containers import ServiceContainer

    container = create_autospec(ServiceContainer)

    with config.patch({SERIALIZER_CONFIG_KEY: DEFAULT_SERIALIZER}):
        container.serializer = config[SERIALIZER_CONFIG_KEY]
        container.accept = [DEFAULT_SERIALIZER]
        yield container


@pytest.fixture(scope='session')
def rabbit_manager(request):
    from nameko.testing import rabbit

    config = request.config
    return rabbit.Client(config.getoption('RABBIT_API_URI'))


@pytest.yield_fixture(scope='session')
def vhost_pipeline(request, rabbit_manager):
    if six.PY2:  # pragma: no cover
        from collections import Iterable  # pylint: disable=E0611
    else:  # pragma: no cover
        from collections.abc import Iterable  # pylint: disable=E0611,E0401
    from six.moves.urllib.parse import urlparse  # pylint: disable=E0401
    import random
    import string
    from kombu.pools import connections
    from nameko.testing.utils import ResourcePipeline
    from nameko.utils.retry import retry

    rabbit_amqp_uri = request.config.getoption('RABBIT_AMQP_URI')
    uri_parts = urlparse(rabbit_amqp_uri)
    username = uri_parts.username

    def create():
        vhost = "nameko_test_{}".format(
            "".join(random.choice(string.ascii_lowercase) for _ in range(10))
        )
        rabbit_manager.create_vhost(vhost)
        rabbit_manager.set_vhost_permissions(
            vhost, username, '.*', '.*', '.*'
        )
        return vhost

    @retry(for_exceptions=Exception, delay=1, max_attempts=9)
    def destroy(vhost):
        rabbit_manager.delete_vhost(vhost)

        # make sure connections for this vhost are also destroyed.
        vhost_pools = [
            pool for key, pool in list(connections.items())
            if isinstance(key, Iterable) and key[4] == vhost
        ]
        for pool in vhost_pools:
            pool.force_close_all()

    pipeline = ResourcePipeline(create, destroy)

    with pipeline.run() as vhosts:
        yield vhosts


@pytest.yield_fixture
def rabbit_uri(request, vhost_pipeline):
    from six.moves.urllib.parse import urlparse  # pylint: disable=E0401

    rabbit_amqp_uri = request.config.getoption('RABBIT_AMQP_URI')
    uri_parts = urlparse(rabbit_amqp_uri)

    with vhost_pipeline.get() as vhost:

        amqp_uri = "{uri.scheme}://{uri.netloc}/{vhost}".format(
            uri=uri_parts, vhost=vhost
        )

        yield amqp_uri


@pytest.yield_fixture
def rabbit_config(rabbit_uri):
    with config.patch({'AMQP_URI': rabbit_uri}):
        yield


@pytest.fixture
def rabbit_ssl_options(request):
    import os
    from test import on_travis

    ssl_options = request.config.getoption('AMQP_SSL_OPTIONS')
    ssl_options = {key: value for key, value in ssl_options} or True

    for file_key in ['ca_certs', 'certfile', 'keyfile']:
        filename = ssl_options.get(file_key)
        if (
                not on_travis and
                filename and not os.path.isfile(filename)
        ):  # pragma: no cover
            pytest.skip('ssl {} file {} does not exist'
                        .format(file_key, filename))

    return ssl_options


@pytest.fixture
def rabbit_ssl_uri(request, rabbit_uri):
    from six.moves.urllib.parse import urlparse  # pylint: disable=E0401

    amqp_ssl_port = request.config.getoption('AMQP_SSL_PORT')
    uri_parts = urlparse(rabbit_uri)
    amqp_ssl_uri = uri_parts._replace(
        netloc=uri_parts.netloc.replace(
            str(uri_parts.port),
            str(amqp_ssl_port))
    ).geturl()

    return amqp_ssl_uri


@pytest.yield_fixture()
def rabbit_ssl_config(request, rabbit_ssl_uri, rabbit_ssl_options):

    conf = {
        'AMQP_URI': rabbit_ssl_uri,
        'AMQP_SSL': rabbit_ssl_options,
    }

    with config.patch(conf):
        yield


@pytest.fixture
def amqp_uri(rabbit_config):
    from nameko.constants import AMQP_URI_CONFIG_KEY

    return config[AMQP_URI_CONFIG_KEY]


@pytest.fixture
def get_message_from_queue(amqp_uri):
    from nameko.amqp import get_connection

    def get(queue_name, ack=True, block=True, timeout=1, accept=None):
        with get_connection(amqp_uri) as conn:
            queue = conn.SimpleQueue(queue_name)

            # queue doesn't allow passing of `accept` to its consumer,
            # so we patch it on after instantiation
            if accept is not None:
                queue.consumer.accept = accept

            message = queue.get(block=block, timeout=timeout)
            if ack:
                message.ack()
            else:
                message.requeue()
            queue.close()
        return message

    return get


@pytest.yield_fixture
def container_factory():
    import nameko
    from nameko.containers import get_container_cls

    all_containers = []

    def make_container(service_cls, config=None):

        # nameko 2.X backward compatible passing of custom config argument
        # we apply config patch if a config dictionary is passed to the factory
        if config:
            patch = nameko.config.patch(config)
            patch.start()
        else:
            patch = None

        container_cls = get_container_cls()
        container = container_cls(service_cls)
        all_containers.append((container, patch))
        return container

    yield make_container

    for container, patch in reversed(all_containers):
        try:
            container.kill()
        except:  # pragma: no cover
            pass
        if patch:
            patch.stop()


@pytest.yield_fixture
def runner_factory():
    import collections
    import nameko
    from nameko.runners import ServiceRunner

    all_runners = []

    def make_runner(*service_classes):

        # nameko 2.X backward compatible passing of custom config argument
        # we apply config patch if a config dictionary is passed to the factory
        if service_classes and isinstance(service_classes[0], collections.Mapping):
            config = service_classes[0]
            service_classes = service_classes[1:]
            patch = nameko.config.patch(config)
            patch.start()
        else:
            patch = None

        runner = ServiceRunner()
        for service_cls in service_classes:
            runner.add_service(service_cls)
        all_runners.append((runner, patch))
        return runner

    yield make_runner

    for runner, patch in reversed(all_runners):
        try:
            runner.kill()
        except:  # pragma: no cover
            pass
        if patch:
            patch.stop()


@pytest.yield_fixture
def predictable_call_ids(request):
    import itertools
    from mock import patch

    with patch('nameko.standalone.rpc.uuid') as client_uuid:
        client_uuid.uuid4.side_effect = (str(i) for i in itertools.count())
        with patch('nameko.containers.uuid', autospec=True) as call_uuid:
            call_uuid.uuid4.side_effect = (str(i) for i in itertools.count())
            yield call_uuid.uuid4


@pytest.yield_fixture
def web_config():
    from nameko.constants import WEB_SERVER_CONFIG_KEY
    from nameko.testing.utils import find_free_port

    port = find_free_port()

    with config.patch({WEB_SERVER_CONFIG_KEY: "127.0.0.1:{}".format(port)}):
        yield


@pytest.fixture()
def web_config_port(web_config):
    from nameko.constants import WEB_SERVER_CONFIG_KEY
    from nameko.web.server import parse_address
    return parse_address(config[WEB_SERVER_CONFIG_KEY]).port


@pytest.yield_fixture()
def web_session(web_config_port):
    from requests import Session
    from werkzeug.urls import url_join

    class WebSession(Session):
        def request(self, method, url, *args, **kwargs):
            url = url_join('http://127.0.0.1:%d/' % web_config_port, url)
            return Session.request(self, method, url, *args, **kwargs)

    sess = WebSession()
    with sess:
        yield sess


@pytest.yield_fixture()
def websocket(web_config_port):
    import eventlet
    from nameko.testing.websocket import make_virtual_socket

    active_sockets = []

    def socket_creator():
        ws_app, wait_for_sock = make_virtual_socket(
            '127.0.0.1', web_config_port)
        gr = eventlet.spawn(ws_app.run_forever)
        active_sockets.append((gr, ws_app))
        socket = wait_for_sock()
        socket.app = ws_app
        return socket

    try:
        yield socket_creator
    finally:
        for gr, ws_app in active_sockets:
            try:
                ws_app.close()
            except Exception:  # pragma: no cover
                pass
            gr.kill()



from __future__ import absolute_import

from time import time

import pytest
from amqp.exceptions import (
    NotFound, PreconditionFailed, RecoverableConnectionError
)
from kombu import Connection
from kombu.common import maybe_declare
from kombu.compression import get_encoder
from kombu.exceptions import OperationalError
from kombu.messaging import Exchange, Producer, Queue
from kombu.serialization import registry
from mock import ANY, MagicMock, Mock, call, patch
from six.moves import queue
from six.moves.urllib.parse import urlparse

from nameko import config
from nameko.amqp.publish import (
    Publisher, UndeliverableMessage, get_connection, get_producer
)
from nameko.constants import AMQP_URI_CONFIG_KEY


def test_get_connection(amqp_uri):
    connection_ids = []

    with get_connection(amqp_uri) as connection:
        connection_ids.append(id(connection))
        assert isinstance(connection, Connection)

    with get_connection(amqp_uri) as connection:
        connection_ids.append(id(connection))
        assert len(set(connection_ids)) == 1


class TestGetProducer(object):

    @pytest.fixture(params=[True, False])
    def confirms(self, request):
        return request.param

    def test_get_producer(self, amqp_uri, confirms):
        producer_ids = []

        with get_producer(amqp_uri, confirms) as producer:
            producer_ids.append(id(producer))
            transport_options = producer.connection.transport_options
            assert isinstance(producer, Producer)
            assert transport_options['confirm_publish'] is confirms

        with get_producer(amqp_uri, confirms) as producer:
            producer_ids.append(id(producer))
            assert len(set(producer_ids)) == 1

    def test_pool_gives_different_producers(self, amqp_uri):
        producer_ids = []

        # get a producer
        with get_producer(amqp_uri, True) as confirmed_producer:
            producer_ids.append(id(confirmed_producer))
            assert len(set(producer_ids)) == 1

        # get a producer with the same parameters
        with get_producer(amqp_uri, True) as confirmed_producer:
            producer_ids.append(id(confirmed_producer))
            assert len(set(producer_ids)) == 1  # same producer returned

        # get a producer with different parameters
        with get_producer(amqp_uri, False) as unconfirmed_producer:
            producer_ids.append(id(unconfirmed_producer))
            assert len(set(producer_ids)) == 2  # different producer returned


class TestPublisherConfirms(object):
    """ Publishing to a non-existent exchange raises if confirms are enabled.
    """

    def test_confirms_disabled(self, amqp_uri):

        with get_producer(amqp_uri, False) as producer:
            producer.publish(
                "msg", exchange="missing", routing_key="key"
            )

    def test_confirms_enabled(self, amqp_uri):

        with pytest.raises(NotFound):
            with get_producer(amqp_uri) as producer:
                producer.publish(
                    "msg", exchange="missing", routing_key="key"
                )


@pytest.mark.filterwarnings("ignore:Mandatory delivery:UserWarning")
@pytest.mark.behavioural
class TestPublisher(object):

    @pytest.fixture
    def routing_key(self):
        return "routing_key"

    @pytest.fixture
    def exchange(self, amqp_uri):
        exchange = Exchange(name="exchange")
        return exchange

    @pytest.fixture
    def queue(self, amqp_uri, exchange, routing_key):
        queue = Queue(
            name="queue", exchange=exchange, routing_key=routing_key
        )
        return queue

    @pytest.fixture
    def publisher(self, amqp_uri, exchange, queue, routing_key):
        return Publisher(
            amqp_uri,
            serializer="json",
            exchange=exchange,
            routing_key=routing_key,
            declare=[exchange, queue]
        )

    def test_routing(
        self, publisher, get_message_from_queue, queue
    ):
        # exchange and routing key can be specified at publish time
        publisher.publish(
            "payload", exchange=None, routing_key=queue.name
        )
        message = get_message_from_queue(queue.name)
        assert message.delivery_info['routing_key'] == queue.name

    @pytest.mark.parametrize("option,value,expected", [
        ('delivery_mode', 1, 1),
        ('priority', 10, 10),
        ('expiration', 10, str(10 * 1000)),
    ])
    def test_delivery_options(
        self, publisher, get_message_from_queue, queue, option, value, expected
    ):
        publisher.publish("payload", **{option: value})
        message = get_message_from_queue(queue.name)
        assert message.properties[option] == expected

    def test_confirms(self, amqp_uri, publisher):
        publisher.mandatory = True

        # confirms are enabled by default;
        # if a mandatory message cannot be delivered, expect an exception
        with pytest.raises(UndeliverableMessage):
            publisher.publish(
                "payload", routing_key="missing", use_confirms=True
            )

        # disabling confirms disables the error
        publisher.publish(
            "payload", routing_key="missing", use_confirms=False
        )

    def test_mandatory_delivery(
        self, publisher, get_message_from_queue, queue
    ):
        # messages are not mandatory by default;
        # no error when routing to a non-existent queue
        publisher.publish("payload", routing_key="missing", mandatory=False)

        # requesting mandatory delivery will result in an exception
        # if there is no bound queue to receive the message
        with pytest.raises(UndeliverableMessage):
            publisher.publish("payload", routing_key="missing", mandatory=True)

        # no exception will be raised if confirms are disabled,
        # even when mandatory delivery is requested,
        # but there will be a warning raised
        with patch('nameko.amqp.publish.warnings') as warnings:
            publisher.publish(
                "payload",
                routing_key="missing",
                mandatory=True,
                use_confirms=False
            )
        assert warnings.warn.called

    @pytest.mark.parametrize("serializer", ['json', 'pickle'])
    def test_serializer(
        self, publisher, get_message_from_queue, queue, serializer
    ):
        payload = {"key": "value"}
        publisher.publish(payload, serializer=serializer)

        content_type, content_encoding, expected_body = (
            registry.dumps(payload, serializer=serializer)
        )

        message = get_message_from_queue(queue.name, accept=content_type)
        assert message.body == expected_body

    def test_compression(
        self, publisher, get_message_from_queue, queue
    ):
        payload = {"key": "value"}
        publisher.publish(payload, compression="gzip")

        _, expected_content_type = get_encoder('gzip')

        message = get_message_from_queue(queue.name)
        assert message.headers['compression'] == expected_content_type

    def test_content_type(
        self, publisher, get_message_from_queue, queue
    ):
        payload = (
            b'GIF89a\x01\x00\x01\x00\x00\xff\x00,'
            b'\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x00;'
        )
        publisher.publish(payload, content_type="image/gif")

        message = get_message_from_queue(queue.name)
        assert message.payload == payload
        assert message.properties['content_type'] == 'image/gif'
        assert message.properties['content_encoding'] == 'binary'

    def test_content_encoding(
        self, publisher, get_message_from_queue, queue
    ):
        payload = "A"
        publisher.publish(
            payload.encode('utf-16'),
            content_type="application/text",
            content_encoding="utf-16"
        )

        message = get_message_from_queue(queue.name)
        assert message.payload == payload
        assert message.properties['content_type'] == 'application/text'
        assert message.properties['content_encoding'] == 'utf-16'

    @pytest.mark.parametrize("headers,extra_headers,expected_headers", [
        # no extra headers
        ({'x-foo': 'foo'}, {}, {'x-foo': 'foo'}),
        # only extra headers
        ({}, {'x-foo': 'foo'}, {'x-foo': 'foo'}),
        # both
        ({'x-foo': 'foo'}, {'x-bar': 'bar'}, {'x-foo': 'foo', 'x-bar': 'bar'}),
        # override
        ({'x-foo': 'foo'}, {'x-foo': 'bar'}, {'x-foo': 'bar'}),
    ])
    @pytest.mark.usefixtures('predictable_call_ids')
    def test_headers(
        self, publisher, get_message_from_queue, queue,
        headers, extra_headers, expected_headers
    ):
        payload = {"key": "value"}

        publisher.publish(
            payload, headers=headers, extra_headers=extra_headers
        )

        message = get_message_from_queue(queue.name)
        assert message.headers == expected_headers
        assert message.properties['application_headers'] == expected_headers

    @pytest.mark.parametrize("option,value,expected", [
        ('reply_to', "queue_name", "queue_name"),
        ('message_id', "msg.1", "msg.1"),
        ('type', "type", "type"),
        ('app_id', "app.1", "app.1"),
        ('cluster_id', "cluster.1", "cluster.1"),
        ('correlation_id', "msg.1", "msg.1"),
    ])
    def test_message_properties(
        self, publisher, get_message_from_queue, queue, option, value, expected
    ):
        publisher.publish("payload", **{option: value})

        message = get_message_from_queue(queue.name)
        assert message.properties[option] == expected

    def test_timestamp(
        self, publisher, get_message_from_queue, queue
    ):
        now = int(time())
        publisher.publish("payload", timestamp=now)

        message = get_message_from_queue(queue.name)
        assert message.properties['timestamp'] == now

    def test_user_id(
        self, publisher, get_message_from_queue, queue, rabbit_config
    ):
        uri_parts = urlparse(config[AMQP_URI_CONFIG_KEY])
        user_id = uri_parts.username

        # successful case
        publisher.publish("payload", user_id=user_id)

        message = get_message_from_queue(queue.name)
        assert message.properties['user_id'] == user_id

        # when user_id does not match the current user, expect an error
        with pytest.raises(PreconditionFailed):
            publisher.publish("payload", user_id="invalid")

    @patch('kombu.messaging.maybe_declare', wraps=maybe_declare)
    def test_declare(
        self, maybe_declare, publisher, get_message_from_queue, routing_key,
        queue, exchange
    ):
        declare = [
            Queue(name="q1", exchange=exchange, routing_key=routing_key),
            Queue(name="q2", exchange=exchange, routing_key=routing_key)
        ]

        publisher.publish("payload", declare=declare)

        assert maybe_declare.call_args_list == [
            call(exchange, ANY, ANY),
            call(queue, ANY, ANY),
            call(declare[0], ANY, ANY),
            call(declare[1], ANY, ANY)
        ]

        assert get_message_from_queue(queue.name).payload == "payload"
        assert get_message_from_queue(declare[0].name).payload == "payload"
        assert get_message_from_queue(declare[1].name).payload == "payload"

    def test_retry(
        self, publisher, get_message_from_queue, rabbit_config
    ):
        mock_publish = MagicMock(__name__="", __doc__="", __module__="")
        mock_publish.side_effect = RecoverableConnectionError("error")

        expected_retries = publisher.retry_policy['max_retries'] + 1

        # with retry
        with patch.object(Producer, '_publish', new=mock_publish):
            with pytest.raises(OperationalError):
                publisher.publish("payload", retry=True)
        assert mock_publish.call_count == 1 + expected_retries

        mock_publish.reset_mock()

        # retry disabled
        with patch.object(Producer, '_publish', new=mock_publish):
            with pytest.raises(RecoverableConnectionError):
                publisher.publish("payload", retry=False)
        assert mock_publish.call_count == 1

    def test_retry_policy(
        self, publisher, get_message_from_queue, rabbit_config
    ):
        mock_publish = MagicMock(__name__="", __doc__="", __module__="")
        mock_publish.side_effect = RecoverableConnectionError("error")

        retry_policy = {
            'max_retries': 5
        }
        expected_retries = retry_policy['max_retries'] + 1

        with patch.object(Producer, '_publish', new=mock_publish):
            with pytest.raises(OperationalError):
                publisher.publish("payload", retry_policy=retry_policy)
        assert mock_publish.call_count == 1 + expected_retries


class TestDefaults(object):

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.parametrize("param", [
        # delivery options
        'delivery_mode', 'mandatory', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy'
    ])
    def test_precedence(self, param, producer):
        """ Verify that a default specified as a class attribute can be
        overridden by a default specified at instantiation time, which can
        further be overridden by a value specified when used.
        """
        publisher_cls = type("Publisher", (Publisher,), {param: "value"})
        publisher = publisher_cls("memory://", **{param: True})

        publisher.publish("payload")
        assert producer.publish.call_args[1][param] is True

        publisher.publish("payload", **{param: False})
        assert producer.publish.call_args[1][param] is False

    def test_header_precedence(self, producer):
        """ Verify that headers at publish time extend any provided
        at instantiation time.
        """
        headers1 = {'h1': Mock()}
        publisher = Publisher("memory://", headers=headers1)

        headers2 = {'h2': Mock()}
        publisher.publish("payload", headers=headers2)

        combined_headers = headers1.copy()
        combined_headers.update(headers2)
        assert producer.publish.call_args[1]["headers"] == combined_headers

        headers3 = {'h3': Mock()}
        publisher.publish("payload", headers=headers3)

        combined_headers = headers1.copy()
        combined_headers.update(headers3)
        assert producer.publish.call_args[1]["headers"] == combined_headers

    def test_declaration_precedence(self, producer):
        """ Verify that declarations at publish time extend any provided
        at instantiation time.
        """
        queue1 = Mock()
        publisher = Publisher("memory://", declare=[queue1])

        queue2 = Mock()
        publisher.publish("payload", declare=[queue2])
        assert producer.publish.call_args[1]["declare"] == [queue1, queue2]

        queue3 = Mock()
        publisher.publish("payload", declare=[queue3])
        assert producer.publish.call_args[1]["declare"] == [queue1, queue3]

    def test_publish_kwargs(self, producer):
        """ Verify that publish_kwargs at publish time augment any provided
        at instantiation time. Verify that publish_kwargs at publish time
        override any provided at instantiation time in the case of a clash.
        Verify that any keyword argument is transparently passed to kombu.
        """
        publisher = Publisher("memory://", reply_to="queue1")
        publisher.publish(
            "payload", reply_to="queue2", correlation_id="1", bogus="bogus"
        )

        # publish-time kwargs override instantiation-time kwargs
        assert producer.publish.call_args[1]["reply_to"] == "queue2"
        # publish-time kwargs augment instantiation-time kwargs
        assert producer.publish.call_args[1]["correlation_id"] == "1"
        # irrelevant keywords pass through transparently
        assert producer.publish.call_args[1]["bogus"] == "bogus"

    def test_use_confirms(self, get_producer):
        """ Verify that publish-confirms can be set as a default specified at
        instantiation time, which can be overridden by a value specified at
        publish time.
        """
        publisher = Publisher("memory://", use_confirms=False)

        publisher.publish("payload")
        use_confirms = get_producer.call_args[0][3].get('confirm_publish')
        assert use_confirms is False

        publisher.publish("payload", use_confirms=True)
        use_confirms = get_producer.call_args[0][3].get('confirm_publish')
        assert use_confirms is True



import pytest
from mock import Mock

from nameko.constants import (
    AUTH_TOKEN_CONTEXT_KEY, LANGUAGE_CONTEXT_KEY, USER_AGENT_CONTEXT_KEY,
    USER_ID_CONTEXT_KEY
)
from nameko.containers import ServiceContainer, WorkerContext
from nameko.contextdata import (
    AuthToken, ContextDataProvider, Language, UserAgent, UserId
)
from nameko.testing.utils import get_extension


CUSTOM_CONTEXT_KEY = "custom"


class CustomValue(ContextDataProvider):
    context_key = CUSTOM_CONTEXT_KEY


class Service(object):
    name = "service"

    # builtin context data dependencies
    auth_token = AuthToken()
    language = Language()
    user_id = UserId()
    user_agent = UserAgent()

    # custom context data dependency
    custom_value = CustomValue()


@pytest.fixture
def container():
    return ServiceContainer(Service)


def test_get_custom_context_value(container):
    dependency = get_extension(
        container, ContextDataProvider, attr_name="custom_value")
    worker_ctx = WorkerContext(
        container, "service", Mock(), data={CUSTOM_CONTEXT_KEY: "hello"})

    assert dependency.get_dependency(worker_ctx) == "hello"


def test_get_unset_value(container):
    dependency = get_extension(
        container, ContextDataProvider, attr_name="custom_value")
    worker_ctx = WorkerContext(
        container, "service", Mock(), data={})

    assert dependency.get_dependency(worker_ctx) is None


@pytest.mark.parametrize('attr_name, context_key', [
    ('auth_token', AUTH_TOKEN_CONTEXT_KEY),
    ('language', LANGUAGE_CONTEXT_KEY),
    ('user_id', USER_ID_CONTEXT_KEY),
    ('user_agent', USER_AGENT_CONTEXT_KEY),

])
def test_get_builtin_dependencies(attr_name, context_key, container):
    dependency = get_extension(
        container, ContextDataProvider, attr_name=attr_name)
    worker_ctx = WorkerContext(
        container, "service", Mock(), data={context_key: 'value'})

    assert dependency.get_dependency(worker_ctx) == "value"



# coding: utf-8
""" Tests for the files and snippets in nameko/docs/examples
"""
import json
import os

import boto3
import jwt
import pytest
from mock import Mock, call, patch
from moto import mock_sqs

from nameko import config
from nameko.exceptions import RemoteError
from nameko.standalone.events import event_dispatcher
from nameko.standalone.rpc import ClusterRpcClient, ServiceRpcClient
from nameko.testing.services import dummy, entrypoint_waiter, entrypoint_hook
from nameko.web.handlers import http


class TestHttp(object):

    def test_http(self, container_factory, web_session):

        from examples.http import HttpService

        container = container_factory(HttpService)
        container.start()

        res = web_session.get("/get/42")
        assert res.status_code == 200
        assert res.text == '{"value": 42}'

        res = web_session.post("/post", data=u"你好".encode('utf-8'))
        assert res.status_code == 200
        assert res.text == u'received: 你好'

        res = web_session.get("/multi")
        assert res.status_code == 200
        assert res.text == 'GET'

        res = web_session.put("/multi")
        assert res.status_code == 200
        assert res.text == 'PUT'

        res = web_session.post("/multi")
        assert res.status_code == 200
        assert res.text == 'POST'

        res = web_session.delete("/multi")
        assert res.status_code == 200
        assert res.text == 'DELETE'

    def test_advanced(self, container_factory, web_session):

        from examples.advanced_http import Service

        container = container_factory(Service)
        container.start()

        res = web_session.get("/privileged")
        assert res.status_code == 403
        assert res.text == 'Forbidden'

        res = web_session.get("/headers")
        assert res.status_code == 201
        assert res.headers['location'] == 'https://www.example.com/widget/1'

        res = web_session.get("/custom")
        assert res.status_code == 200
        assert res.text == 'payload'

    def test_custom_exception(self, container_factory, web_session):

        from examples.http_exceptions import Service

        container = container_factory(Service)
        container.start()

        res = web_session.get("/custom_exception")
        assert res.status_code == 400
        assert res.headers['Content-Type'] == 'application/json'
        assert res.json() == {
            'error': 'INVALID_ARGUMENTS',
            'message': "Argument `foo` is required.",
        }

    def test_will_not_handle_unknown_exception(self, container_factory, web_session):

        from examples.http_exceptions import http

        class Service(object):
            name = "service"

            @http('GET', '/exception')
            def exception(self, request):
                raise ValueError("Argument `foo` is required.")

        container = container_factory(Service)
        container.start()

        res = web_session.get("/exception")
        assert res.status_code == 500


@pytest.mark.usefixtures("rabbit_config")
class TestEvents(object):

    def test_events(self, container_factory):

        from examples.events import ServiceA, ServiceB

        container_a = container_factory(ServiceA)
        container_b = container_factory(ServiceB)
        container_a.start()
        container_b.start()

        with ServiceRpcClient('service_a') as service_a_rpc:

            with patch.object(ServiceB, 'handle_event') as handle_event:

                with entrypoint_waiter(container_b, 'handle_event'):
                    service_a_rpc.dispatching_method("event payload")
                assert handle_event.call_args_list == [call("event payload")]

            # test without the patch to catch any errors in the handler method
            with entrypoint_waiter(container_b, 'handle_event'):
                service_a_rpc.dispatching_method("event payload")

    def test_standalone_events(self, container_factory):

        from examples.events import ServiceB

        container_b = container_factory(ServiceB)
        container_b.start()

        # standalone example doesn't import due to undefined variables
        # use execfile with a local namespace
        ns = {'AMQP_URI': config['AMQP_URI']}

        dirpath = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        filepath = os.path.join(dirpath, 'standalone_events.py')

        with entrypoint_waiter(container_b, 'handle_event'):
            with open(filepath) as f:
                code = compile(f.read(), filepath, 'exec')
                exec(code, globals(), ns)

    def test_event_broadcast(self, container_factory):

        from examples.event_broadcast import ListenerService

        container_1 = container_factory(ListenerService)
        container_2 = container_factory(ListenerService)
        container_1.start()
        container_2.start()

        dispatch = event_dispatcher()

        with patch.object(ListenerService, 'ping') as ping:

            waiter_1 = entrypoint_waiter(container_1, 'ping')
            waiter_2 = entrypoint_waiter(container_2, 'ping')

            with waiter_1, waiter_2:
                dispatch("monitor", "ping", "payløad")
            assert ping.call_count == 2

        # test without the patch to catch any errors in the handler method
        with entrypoint_waiter(container_1, 'ping'):
            dispatch("monitor", "ping", "payløad")


@pytest.mark.usefixtures("rabbit_config")
class TestAnatomy(object):

    def test_anatomy(self, container_factory):

        from examples.anatomy import Service

        container = container_factory(Service)
        container.start()

        with ServiceRpcClient('service') as service_rpc:
            assert service_rpc.method() is None


@pytest.mark.usefixtures("rabbit_config")
class TestHelloWorld(object):

    def test_hello_world(self, container_factory):

        from examples.helloworld import GreetingService

        container = container_factory(GreetingService)
        container.start()

        with ServiceRpcClient('greeting_service') as greet_rpc:
            assert greet_rpc.hello(u"Møtt") == u"Hello, Møtt!"


@pytest.mark.usefixtures("rabbit_config")
class TestRpc(object):

    def test_rpc(self, container_factory):

        from examples.rpc import ServiceX, ServiceY

        container_x = container_factory(ServiceX)
        container_y = container_factory(ServiceY)
        container_x.start()
        container_y.start()

        with ServiceRpcClient('service_x') as service_x_rpc:
            assert service_x_rpc.remote_method(u"føø") == u"føø-x-y"

    def test_standalone_rpc(self, container_factory):

        from examples.rpc import ServiceX, ServiceY

        container_x = container_factory(ServiceX)
        container_y = container_factory(ServiceY)
        container_x.start()
        container_y.start()

        # standalone example doesn't import due to undefined variables
        # use execfile with a local namespace
        ns = {'AMQP_URI': config['AMQP_URI']}

        dirpath = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        filepath = os.path.join(dirpath, 'standalone_rpc.py')

        with entrypoint_waiter(container_x, 'remote_method'):
            with open(filepath) as f:
                code = compile(f.read(), filepath, 'exec')
                exec(code, globals(), ns)

    def test_async_rpc(self, container_factory):

        from examples.rpc import ServiceX, ServiceY

        container_x = container_factory(ServiceX)
        container_y = container_factory(ServiceY)
        container_x.start()
        container_y.start()

        # async example doesn't import due to undefined variables
        # use execfile with a local namespace
        ns = {
            'config': config,
            'ClusterRpcClient': ClusterRpcClient
        }

        dirpath = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        filepath = os.path.join(dirpath, 'async_rpc.py')

        with entrypoint_waiter(container_x, 'remote_method'):
            with open(filepath) as f:
                code = compile(f.read(), filepath, 'exec')
                exec(code, globals(), ns)


class TestServiceContainer(object):

    def test_service_container(self):
        from examples import service_container


class TestServiceRunner(object):

    def test_service_runner(self):
        from examples import service_runner


@pytest.mark.usefixtures("rabbit_config")
class TestTimer(object):

    def test_timer(self, container_factory):

        from examples.timer import Service

        container = container_factory(Service)

        with entrypoint_waiter(container, 'ping'):
            container.start()


@pytest.mark.usefixtures("rabbit_config")
class TestTravis(object):

    @pytest.fixture
    def fake_travis(self, container_factory, web_config):

        class FakeTravis:
            name = "travis"

            @http("GET", "/fake_status")
            def method(self, request):
                return json.dumps({
                    "last_build_result": "success",
                    "slug": "nameko/nameko",
                    "last_build_finished_at": "2020-01-01"
                })

        container = container_factory(FakeTravis)
        container.start()

        fake_url = "http://{}/fake_status".format(config['WEB_SERVER_ADDRESS'])
        with patch("examples.travis.URL_TEMPLATE", new=fake_url):
            yield

    @pytest.mark.usefixtures("fake_travis")
    def test_travis(self, container_factory):

        from examples.travis import Travis

        container = container_factory(Travis)
        container.start()

        with ServiceRpcClient('travis_service') as travis_rpc:
            status = travis_rpc.status_message("nameko", "nameko")
            assert "Project nameko/nameko" in status


@pytest.mark.usefixtures("web_config")
class TestWebsocketRpc(object):

    def test_websocket_rpc(self, container_factory, websocket):

        from examples.websocket_rpc import WebsocketRpc

        container = container_factory(WebsocketRpc)
        container.start()

        ws = websocket()
        assert ws.rpc('echo', value=u"hellø") == u'hellø'


class TestAuth:

    @pytest.fixture
    def db(self):
        return {
            'matt': {
                'password': (
                    b'$2b$12$fZXR7Z1Eoyn0pfym8.'
                    b'LyRuIFabYj00ZzhdaJ0qoTLZs9w4fg3pKlK'
                ),
                'roles': [
                    'developer',
                ]
            },
            'susie': {
                'password': (
                    b'$2b$12$k4MVi9PcbSsOqONoj5vW9.'
                    b'pcQpB0xSjYkZcc6Ogr5nQ4MD8DRDiUK'
                ),
                'roles': [
                    'developer',
                    'admin'
                ]
            }
        }

    def test_authenticate(self, db):
        from examples.auth import Auth, JWT_SECRET

        worker_ctx = Mock(context_data={})
        dep = Auth.Api(db, worker_ctx)
        token = dep.authenticate("matt", "secret")
        jwt.decode(token, key=JWT_SECRET, verify=True)
        assert worker_ctx.context_data['auth'] == token

    def test_authenticate_bad_username(self, db):
        from examples.auth import Auth, Unauthenticated

        worker_ctx = Mock(context_data={})
        dep = Auth.Api(db, worker_ctx)
        with pytest.raises(Unauthenticated):
            dep.authenticate("angela", "secret")
        assert worker_ctx.context_data.get('auth') is None

    def test_authenticate_bad_password(self, db):
        from examples.auth import Auth, Unauthenticated

        worker_ctx = Mock(context_data={})
        dep = Auth.Api(db, worker_ctx)
        with pytest.raises(Unauthenticated):
            dep.authenticate("matt", "invalid")
        assert worker_ctx.context_data.get('auth') is None

    def test_authenticated_user_has_role(self, db):
        from examples.auth import Auth, JWT_SECRET

        token = jwt.encode(
            {'username': 'matt', 'roles': ['dev']}, key=JWT_SECRET
        )

        worker_ctx = Mock(context_data={'auth': token})
        dep = Auth.Api(db, worker_ctx)
        assert dep.has_role('dev')

    def test_authenticated_user_does_not_have_role(self, db):
        from examples.auth import Auth, JWT_SECRET

        token = jwt.encode(
            {'username': 'matt', 'roles': ['dev']}, key=JWT_SECRET
        )

        worker_ctx = Mock(context_data={'auth': token})
        dep = Auth.Api(db, worker_ctx)
        assert not dep.has_role('admin')

    def test_has_role_unauthenicated_user(self, db):
        from examples.auth import Auth, Unauthenticated

        worker_ctx = Mock(context_data={})
        dep = Auth.Api(db, worker_ctx)
        with pytest.raises(Unauthenticated):
            dep.has_role('admin')

    def test_has_role_invalid_token(self, db):
        from examples.auth import Auth

        worker_ctx = Mock(context_data={'auth': 'invalid-token'})
        dep = Auth.Api(db, worker_ctx)
        assert not dep.has_role('admin')

    def test_check_role(self, db):
        from examples.auth import Auth, Unauthorized, JWT_SECRET

        token = jwt.encode(
            {'username': 'matt', 'roles': ['dev']}, key=JWT_SECRET
        )

        worker_ctx = Mock(context_data={'auth': token})
        dep = Auth.Api(db, worker_ctx)

        assert dep.check_role('dev') is None
        with pytest.raises(Unauthorized):
            dep.check_role('admin')


@pytest.mark.usefixtures("rabbit_config")
class TestExpectedExceptions:

    def test_expected_exceptions(self, container_factory):
        from examples.auth import JWT_SECRET
        from examples.expected_exceptions import Service

        container = container_factory(Service)
        container.start()

        token = jwt.encode({"roles": []}, key=JWT_SECRET)

        with ServiceRpcClient(
            "service", context_data={"auth": token}
        ) as client:
            with pytest.raises(RemoteError) as exc:
                client.update(None)
            assert exc.value.exc_type == 'Unauthorized'

        admin_token = jwt.encode({"roles": ['admin']}, key=JWT_SECRET)

        with ServiceRpcClient(
            "service", context_data={"auth": admin_token}
        ) as client:
            with pytest.raises(RemoteError) as exc:
                client.update(None)
            assert exc.value.exc_type == 'TypeError'


@pytest.mark.usefixtures("rabbit_config")
class TestSensitiveArguments:

    def test_sensitive_arguments(self, container_factory):
        from examples.auth import JWT_SECRET
        from examples.sensitive_arguments import Service

        container = container_factory(Service)
        container.start()

        with ServiceRpcClient("service") as client:
            token = client.login("matt", "secret")
            jwt.decode(token, key=JWT_SECRET, verify=True)
            with pytest.raises(RemoteError) as exc:
                client.login("matt", "incorrect")
            assert exc.value.exc_type == "Unauthenticated"


class TestSqsReceive:

    @pytest.yield_fixture
    def sqs_client(self):
        with mock_sqs():
            client = boto3.client('sqs', region_name="eu-west-1")
            yield client

    @pytest.yield_fixture
    def queue(self, sqs_client):
        queue = sqs_client.create_queue(QueueName="nameko-sqs")
        url = queue['QueueUrl']
        yield url
        sqs_client.delete_queue(QueueUrl=url)

    def test_sqs_receive(self, queue, sqs_client, container_factory):

        from examples.sqs_receive import receive

        class Service:
            name = "sqs-service"

            @receive(queue)
            def handle(self, payload):
                return payload

        container = container_factory(Service)
        container.start()

        with entrypoint_waiter(container, 'handle') as res:
            sqs_client.send_message(QueueUrl=queue, MessageBody="hello")
        assert res.get() == "hello"


class TestSqsSend:

    @pytest.yield_fixture
    def sqs_client(self):
        with mock_sqs():
            client = boto3.client('sqs', region_name="eu-west-1")
            yield client

    @pytest.yield_fixture
    def queue(self, sqs_client):
        queue = sqs_client.create_queue(QueueName="nameko-sqs")
        url = queue['QueueUrl']
        yield url
        sqs_client.delete_queue(QueueUrl=url)

    def test_sqs_send(self, queue, sqs_client, container_factory):

        from examples.sqs_send import SqsSend

        class Service:
            name = "sqs-service"

            send_message = SqsSend(queue)

            @dummy
            def method(self, payload):
                self.send_message(payload)

        container = container_factory(Service)
        container.start()

        with entrypoint_hook(container, 'method') as send:
            send("hello")

        resp = sqs_client.receive_message(QueueUrl=queue)
        assert resp['Messages'][0]['Body'] == "hello"


class TestSqsService:

    @pytest.yield_fixture
    def sqs_client(self):
        with mock_sqs():
            client = boto3.client('sqs', region_name="eu-west-1")
            yield client

    @pytest.yield_fixture
    def queue(self, sqs_client):
        queue = sqs_client.create_queue(QueueName="nameko-sqs")
        url = queue['QueueUrl']
        yield url
        sqs_client.delete_queue(QueueUrl=url)

    def test_sqs_service(self, queue, sqs_client, container_factory):

        from examples.sqs_service import SqsService

        container = container_factory(SqsService)
        container.start()

        with entrypoint_waiter(container, 'handle_sqs_message') as res:
            sqs_client.send_message(QueueUrl=queue, MessageBody="hello")
        assert res.get() == "hello"



import click

from nameko import config
from nameko.exceptions import CommandError, ConfigurationError

from .utils import setup_config
from .click_arguments import argument_services
from .click_options import (
    option_broker,
    option_config_file,
    option_define,
    option_backdoor_port,
    option_interface,
)


# main nameko command
@click.group()
@click.version_option()
def cli():
    """CLI to manage nameko based services and configuration.
    """
    pass


# nameko show-config
@cli.command()
@option_config_file()
@option_define()
def show_config(config_file, define):
    """Output as YAML string the configuration that would be passed to a service.
    Useful for viewing config files that load values from environement variables.
    """
    import yaml

    setup_config(config_file, define)
    click.echo(yaml.dump(config.data))


# nameko run
@cli.command()
@option_broker()
@option_backdoor_port()
@option_config_file()
@option_define()
@argument_services()
def run(broker, config_file, define, backdoor_port, services):
    """Run nameko services. Given a python path to a module containing one or more
nameko services, will host and run them. By default this will try to find
classes that look like services (anything with nameko entrypoints), but a
specific service can be specified via ``nameko run module:ServiceClass``.

positional arguments:

  module[:service class]

        python path to one or more service classes to run
    """

    from .run import main

    try:
        setup_config(config_file, define, broker)
        if backdoor_port:
            host, port = backdoor_port
            msg = "To connect to backdoor: `$ telnet {host} {port}`"
            click.echo(msg.format(host=host, port=port))
        main(services, backdoor_port)
    except (CommandError, ConfigurationError) as exc:
        click.echo("Error: {}".format(exc))
        raise click.Abort()


# nameko shell
@cli.command()
@option_broker()
@option_interface()
@option_config_file()
@option_define()
def shell(broker, interface, config_file, define):
    from .shell import main

    setup_config(config_file, define, broker)
    main(interface)
